import pandas as pd
import numpy as np
import math

def round_to_sf(x, sf=5):
 if x==0:
  return 0
 else:
  return round(x,sf-1-int(math.floor(math.log10(abs(x)))))

def get_gradient(x, y):
 aveX = float(sum(x))/len(x)
 aveY = float(sum(y))/len(y)
 
 numerator=sum((x-aveX)*(y-aveY))
 denominator=sum((x-aveX)*(x-aveX))
 
 return numerator/denominator


def get_sorted_keys(d):
 keys = [c for c in d]
 keys.sort(key=str)
 return keys


import numpy as np
import pandas as pd

from . import util

def produce_cont_relevances(inputSeries, cont):
 reles=np.zeros((len(cont)+1,len(inputSeries)))
 
 reles[0][(inputSeries<=cont[0][0])] = 1 #d(featpred)/d(pt)
 for i in range(len(cont)-1):
  x = inputSeries
  x1 = cont[i][0]
  x2 = cont[i+1][0]
  subset = (x>=x1) & (x<=x2)
  reles[i][subset] = (x2 - x[subset])/(x2 - x1) #d(featpred)/d(pt)
  reles[i+1][subset] = (x[subset] - x1)/(x2 - x1) #d(featpred)/d(pt)
 reles[-2][(inputSeries>=cont[-1][0])] = 1 #d(featpred)/d(pt)
 
 reles[-1] = 1 - np.sum(reles, axis=0)
 return np.transpose(reles) #roundabout way of doing this but the rest of the function doesn't flow as naturally if x and y don't switch places

def produce_cont_relevances_dict(inputDf, model):
 opDict = {}
 
 if "conts" in model:
  for col in model["conts"]:
   opDict[col]=produce_cont_relevances(inputDf[col], model['conts'][col])
 
 return opDict

def produce_cat_relevances(inputSeries, cat):
 reles=np.zeros((len(cat["uniques"])+1,len(inputSeries)))
 
 skeys = util.get_sorted_keys(cat["uniques"])
 for i in range(len(skeys)):
  reles[i][inputSeries.isin([skeys[i]])] = 1 #d(featpred)/d(pt)
 reles[-1][~inputSeries.isin(skeys)] = 1 #d(featpred)/d(pt)
 
 return np.transpose(reles) #roundabout way of doing this but the rest of the function doesn't flow as naturally if x and y don't switch places

def produce_cat_relevances_dict(inputDf, model):
 opDict = {}
 
 if "cats" in model:
  for col in model["cats"]:
   opDict[col]=produce_cat_relevances(inputDf[col], model['cats'][col])
 
 return opDict

#Interactions

def interact_relevances(relesA, relesB):
 
 relesA = np.transpose(relesA) #Yes, I know.
 relesB = np.transpose(relesB) #Shut up.
 
 relesI = np.zeros((len(relesA)*len(relesB),len(relesA[0])))
 
 for i in range(len(relesA)):
  for j in range(len(relesB)):
   relesI[i*len(relesB)+j] = relesA[i]*relesB[j]
 
 return np.transpose(relesI) # . . .

def produce_catcat_relevances(inputSeries1, inputSeries2, catcat):
 return interact_relevances(produce_cat_relevances(inputSeries1, catcat), produce_cat_relevances(inputSeries2, catcat["OTHER"]))

def produce_catcont_relevances(inputSeries1, inputSeries2, catcont):
 return interact_relevances(produce_cat_relevances(inputSeries1, catcont), produce_cont_relevances(inputSeries2, catcont["OTHER"]))

def produce_contcont_relevances(inputSeries1, inputSeries2, contcont):
 return interact_relevances(produce_cont_relevances(inputSeries1, contcont), produce_cont_relevances(inputSeries2, contcont[0][1]))

def produce_interxn_relevances_dict(inputDf, model):
 
 opDict = {}
 if 'catcats' in model:
  for cols in model['catcats']:
   col1, col2 = cols.split(' X ')
   opDict[cols] = produce_catcat_relevances(inputDf[col1], inputDf[col2], model['catcats'][cols])
 
 if 'catconts' in model:
  for cols in model['catconts']:
   col1, col2 = cols.split(' X ')
   opDict[cols] = produce_catcont_relevances(inputDf[col1], inputDf[col2], model['catconts'][cols])
 
 if 'contconts' in model:
  for cols in model['contconts']:
   col1, col2 = cols.split(' X ')
   opDict[cols] = produce_contcont_relevances(inputDf[col1], inputDf[col2], model['contconts'][cols])
 
 return opDict






def sum_and_listify_matrix(a):
 return np.array(sum(a)).tolist()

def produce_total_irelevances_dict(releDict):
 op = {}
 for cols in releDict:
  op[cols] = sum_and_listify_matrix(releDict[cols])
 return op

def produce_total_relevances_dict(contReleDict, catReleDict):
 op = {"conts":{},"cats":{}}
 for col in contReleDict:
  op["conts"][col] = sum_and_listify_matrix(contReleDict[col])
 for col in catReleDict:
  op["cats"][col] = sum_and_listify_matrix(catReleDict[col])
 return op



def produce_wReleDict(releDict, w):
 wReleDict = {}
 for col in releDict:
  wReleDict[col]=w*releDict[col]
 return wReleDict



import pandas as pd
import numpy as np
import math
import copy

def prep_model(df, resp, cats, conts, catMinPrev=0.01, contTargetPts=5, contEdge=0.01, defaultValue=1, weightCol=None):
 
 if weightCol==None:
  model={"BASE_VALUE":df[resp].mean(), "featcomb":"mult"}
 else:
  model={"BASE_VALUE":sum(df[resp]*df[weightCol])/sum(df[weightCol]), "featcomb":"mult"}
 
 for cat in cats:
  model = add_cat_to_model(model, df, cat, catMinPrev, defaultValue, weightCol)
 for cont in conts:
  model = add_cont_to_model(model, df, cont, contTargetPts,contEdge, defaultValue, weightCol)
 
 return model


def get_cat_feat(df, cat, catMinPrev=0.01, defaultValue=1, weightCol=None):
 
 if weightCol==None:
  df["WEIGHT_COL"]=1
 else:
  df["WEIGHT_COL"] = df[weightCol]
 
 sw = sum(df["WEIGHT_COL"])
 
 feat = {"OTHER":defaultValue, "uniques":{}}
 
 for u in df[cat].unique():
  if sum(df[df[cat]==u]["WEIGHT_COL"]/sw)>catMinPrev:
   feat['uniques'][u]=defaultValue
 
 return feat

def add_cat_to_model(model, df, cat, catMinPrev=0.01, defaultValue=1, weightCol=None):
 
 if "cats" not in model:
  model["cats"]={}
 model['cats'][cat] = get_cat_feat(df,cat,catMinPrev, defaultValue, weightCol)
 
 return model



def get_cont_feat(df, cont, contTargetPts=5, contEdge=0.01, defaultValue=1, weightCol=None):
 
 if weightCol==None:
  df["WEIGHT_COL"]=1
 else:
  df["WEIGHT_COL"] = df[weightCol]
 
 df = df[~df[cont].isna()]
 
 sw = sum(df["WEIGHT_COL"])
 
 inpts = []
 for i in range(contTargetPts):
  inpts.append(contEdge+(1-contEdge*2)*i/(contTargetPts-1))
 
 df = df.sort_values(by=cont).reset_index()
 df["CSWC"]=df["WEIGHT_COL"].cumsum()
 
 pts=[]
 feat=[]
 
 for inpt in inpts:
  newpt = min(df[df["CSWC"]>=(inpt*sw)][cont])
  if newpt not in pts:
   pts.append(newpt)
 
 for pt in pts:
  feat.append([pt,defaultValue])
 
 return feat

def add_cont_to_model(model, df, cont, contTargetPts=5, contEdge=0.01, defaultValue=1, weightCol=None):

 if "conts" not in model:
  model["conts"]={}
 model['conts'][cont] = get_cont_feat(df,cont,contTargetPts, contEdge, defaultValue, weightCol)
 
 return model



def get_contcont_feat(df, cont1, cont2, contTargetPts1=3, contTargetPts2=3, contEdge1=0.05, contEdge2=0.05, defaultValue=1, weightCol=None):
 
 feat1 = get_cont_feat(df, cont1, contTargetPts1, contEdge1, defaultValue, weightCol)
 feat2 = get_cont_feat(df, cont2, contTargetPts2, contEdge2, defaultValue, weightCol)
 
 for i in range(len(feat1)):
  feat1[i][1]=copy.deepcopy(feat2)
 
 return feat1

def add_contcont_to_model(model, df, cont1, cont2, contTargetPts1=3, contTargetPts2=3, contEdge1=0.05, contEdge2=0.05, defaultValue=1, weightCol=None, replace=False):
 
 if "contconts" not in model:
  model["contconts"]={}
 model['contconts'][cont1 + " X " + cont2] = get_contcont_feat(df, cont1, cont2, contTargetPts1, contTargetPts2, contEdge1, contEdge2, defaultValue, weightCol)
 
 if replace:
  if "conts" in model:
   if cont1 in model["conts"]:
    del model["conts"][cont1]
   if cont2 in model["conts"]:
    del model["conts"][cont2]
 
 return model



def get_catcont_feat(df, cat, cont, catMinPrev=0.1, contTargetPts=3, contEdge=0.05, defaultValue=1, weightCol=None):
 
 feat1 = get_cat_feat(df, cat, catMinPrev, defaultValue, weightCol)
 feat2 = get_cont_feat(df, cont, contTargetPts, contEdge, defaultValue, weightCol)
 
 feat1["OTHER"]=copy.deepcopy(feat2)
 
 for u in feat1["uniques"]:
  feat1["uniques"][u]=copy.deepcopy(feat2)
 
 return feat1

def add_catcont_to_model(model, df, cat, cont, catMinPrev=0.1, contTargetPts=3, contEdge=0.05, defaultValue=1, weightCol=None, replace=False):
 
 if "catconts" not in model:
  model["catconts"]={}
 model['catconts'][cat + " X " + cont] = get_catcont_feat(df, cat, cont, catMinPrev, contTargetPts, contEdge, defaultValue, weightCol)
 
 if replace:
  if "cats" in model:
   if cat in model["cats"]:
    del model["cats"][cat]
  if "conts" in model:
   if cont in model["conts"]:
    del model["conts"][cont]
 
 return model



def get_catcat_feat(df, cat1, cat2, catMinPrev1=0.1, catMinPrev2=0.1, defaultValue=1, weightCol=None):
 
 feat1 = get_cat_feat(df, cat1, catMinPrev1, defaultValue, weightCol)
 feat2 = get_cat_feat(df, cat2, catMinPrev2, defaultValue, weightCol)
 
 feat1["OTHER"]=copy.deepcopy(feat2)
 
 for u in feat1["uniques"]:
  feat1["uniques"][u]=copy.deepcopy(feat2)
 
 return feat1

def add_catcat_to_model(model, df, cat1, cat2, catMinPrev1=0.1, catMinPrev2=0.1, defaultValue=1, weightCol=None, replace=False):
 
 if "catcats" not in model:
  model["catcats"]={}
 model['catcats'][cat1 + " X " + cat2] = get_catcat_feat(df, cat1, cat2, catMinPrev1, catMinPrev2, defaultValue, weightCol)
 
 if replace:
  if "cats" in model:
   if cat1 in model["cats"]:
    del model["cats"][cat1]
   if cat2 in model["cats"]:
    del model["cats"][cat2]
 
 return model


if __name__ == '__main__':
 df = pd.DataFrame({"cat1":['a','a','a','b','b','b','q','q','q','q','q'],'cont1':[1,2,3,1,2,3,1,2,3,1.5,np.nan],"y":[1,2,3,4,5,6,7,8,9,10,11]})
 print(get_cont_feat(df, "cont1", 5, 0.001, 1))
 
 



import copy
import pandas as pd

from . import misc
from . import util

exampleModel = {"BASE_VALUE":1700,"conts":{"cont1":[[0.01, 1],[0.02,1.1], [0.03, 1.06]], "cont2":[[37,1.2],[98, 0.9]]}, "cats":{"cat1":{"uniques":{"wstfgl":1.05, "florpalorp":0.92}, "OTHER":1.04}}}

def uniquify_model(model, df):
 newModel = copy.deepcopy(model)
 
 if "conts" in model:
  for col in model["conts"]:
   uList = df[col].unique()
   uList = [u for u in uList]
   uList.sort()
   newModel["conts"][col] = [[u,misc.get_effect_of_this_cont_on_single_input(u, model["conts"][col])] for u in uList]
 
 if "cats" in model:
  for col in model["cats"]:
   uList = df[col].unique()
   uList = [u for u in uList]
   uList.sort(key=str)
   newModel["cats"][col]={"uniques":{}, "OTHER":model["cats"][col]["OTHER"]}
   for u in uList:
    newModel["cats"][col]["uniques"][u] = misc.get_effect_of_this_cat_on_single_input(u, model["cats"][col])
 
 if "catcats" in model:
  for cols in model["catcats"]:
   cat1, cat2 = cols.split(" X ")
   
   uList1 = df[cat1].unique()
   uList1 = [u for u in uList1]
   uList1.sort(key=str)
   
   uList2 = df[cat2].unique()
   uList2 = [u for u in uList2]
   uList2.sort(key=str)
   
   newModel["catcats"][cols]={"uniques":{}, "OTHER":{"uniques":{}, "OTHER":model["catcats"][cols]["OTHER"]["OTHER"]}}
   
   for u2 in uList2:
    newModel["catcats"][cols]["OTHER"]["uniques"][u2] = misc.get_effect_of_this_catcat_on_single_input("SolidGoldMagikarp", u2, model["catcats"][cols])
   
   for u1 in uList1:
    newModel["catcats"][cols]["uniques"][u1] = {"uniques":{},"OTHER":misc.get_effect_of_this_catcat_on_single_input(u1,"SolidGoldMagikarp", model["catcats"][cols])}
    for u2 in uList2:
     newModel["catcats"][cols]["uniques"][u1]["uniques"][u2] = misc.get_effect_of_this_catcat_on_single_input(u1,u2, model["catcats"][cols])
   
  if "catconts" in model:
   for cols in model["catconts"]:
    cat, cont = cols.split(" X ")
    
    uList1 = df[cat].unique()
    uList1 = [u for u in uList1]
    uList1.sort(key=str)
    
    uList2 = df[cont].unique()
    uList2 = [u for u in uList2]
    uList2.sort()
    
    newModel["catconts"][cols] = {"uniques":{}, "OTHER":[[u2, misc.get_effect_of_this_catcont_on_single_input("SolidGoldMagikarp", u2, model["catconts"][cols])] for u2 in uList2]}
    for u1 in uList1:
     newModel["catconts"][cols]["uniques"][u1] = [[u2, misc.get_effect_of_this_catcont_on_single_input(u1, u2, model["catconts"][cols])] for u2 in uList2]
  
  if "contconts" in model:
   for cols in model["contconts"]:
    cont1, cont2 = cols.split(" X ")
    
    uList1 = df[cont1].unique()
    uList1 = [u for u in uList1]
    uList1.sort()
    
    uList2 = df[cont2].unique()
    uList2 = [u for u in uList2]
    uList2.sort()
    
    newModel["contconts"][cols]=[]
    for u1 in uList1:
     newModel["contconts"][cols].append([u1, [[u2, misc.get_effect_of_this_contcont_on_single_input(u1,u2,model["contconts"][cols])] for u2 in uList2]])
 
 return newModel

def find_max_len(model, multip=1):
 ml=0
 for col in model['conts']:
  if (len(model['conts'][col])*multip)>ml:
   ml = (len(model['conts'][col])*multip)
 for col in model['cats']:
  if (len(model['cats'][col]['uniques'])+1)>ml:
   ml = len(model['cats'][col]['uniques'])+1
 return ml

def get_cont_inputs(cont, detail=1):
 op=[]
 for i in range(len(cont)-1):
  for j in range(detail):
   op.append(float(cont[i][0]*(detail-j) + cont[i+1][0]*j)/detail)
 op.append(cont[-1][0])
 return op

def linesify_catcat(catcat):
 
 skeys1 = util.get_sorted_keys(catcat["uniques"])
 skeys2 = util.get_sorted_keys(catcat["OTHER"]["uniques"])
 
 op=[',']*(len(skeys2)+2)
 
 #Headline
 op[0]=op[0]+','
 for k1 in skeys1:
  op[0] = op[0]+","+k1
 op[0] = op[0]+",OTHER"
 
 #Work downwards
 for i in range(len(skeys2)):
  op[i+1]=op[i+1]+','+skeys2[i]
  for j in range(len(skeys1)):
   op[i+1] = op[i+1]+','+str(misc.get_effect_of_this_catcat_on_single_input(skeys1[j],skeys2[i],catcat))#str(catcat['uniques'][skeys1[j]]['uniques'][skeys2[i]])
  op[i+1] = op[i+1]+','+str(misc.get_effect_of_this_catcat_on_single_input("OTHER",skeys2[i],catcat))
 
 op[-1]=op[-1]+",OTHER"
 for j in range(len(skeys1)):
  op[-1]=op[-1]+','+str(misc.get_effect_of_this_catcat_on_single_input(skeys1[j],"OTHER",catcat))
 op[-1]=op[-1]+','+str(misc.get_effect_of_this_catcat_on_single_input("OTHER","OTHER",catcat))
 
 return op

def linesify_catcont(catcont, detail=1):
 
 skeys = util.get_sorted_keys(catcont["uniques"])
 contPuts=get_cont_inputs(catcont['OTHER'], detail)
 
 op=[',']*(len(contPuts)+2)
 
 #Headline
 op[0]=op[0]+','
 for k in skeys:
  op[0] = op[0]+","+k
 op[0] = op[0]+",OTHER"
 
 #Work downwards
 for i in range(len(contPuts)):
  op[i+1]=op[i+1]+','+str(contPuts[i])
  for j in range(len(skeys)):
   op[i+1] = op[i+1]+','+str(misc.get_effect_of_this_catcont_on_single_input(skeys[j],contPuts[i], catcont))
  op[i+1] = op[i+1]+','+str(misc.get_effect_of_this_catcont_on_single_input("OTHER",contPuts[i], catcont))
 
 return op

def linesify_contcont(contcont, detail=1):
 
 contPuts1=get_cont_inputs(contcont, detail)
 contPuts2=get_cont_inputs(contcont[0][1], detail)
 
 op=[',']*(len(contPuts2)+2)
 
 #Headline
 op[0]=op[0]+','
 for contPut in contPuts1:
  op[0] = op[0]+","+str(contPut)
 
 #Work downwards
 for i in range(len(contPuts2)):
  op[i+1]=op[i+1]+','+str(contPuts2[i])
  for j in range(len(contPuts1)):
   op[i+1] = op[i+1]+','+str(misc.get_effect_of_this_contcont_on_single_input(contPuts1[j],contPuts2[i], contcont))
 
 return op

def export_model(model, detail=1, filename="op.csv"):
 lines = ['']*(find_max_len(model, detail)+4)
 
 #Add base value
 
 for l in range(len(lines)):
  if l==1:
   lines[l] = lines[l]+",BASE_VALUE"
  elif l==2:
   lines[l] = lines[l]+","+str(model['BASE_VALUE'])
  else:
   lines[l] = lines[l]+','
 
 #Add conts
 
 if 'conts' in model:
  for col in model['conts']:
   lines[0] = lines[0]+',,,'
   lines[1] = lines[1]+',,'+col+','
   contPuts=get_cont_inputs(model['conts'][col], detail)
   for i in range(len(contPuts)):
     lines[i+2] = lines[i+2] + ',,' + str(contPuts[i]) + ',' + str(misc.get_effect_of_this_cont_on_single_input(contPuts[i], model['conts'][col]))
   for l in range(len(contPuts)+2, len(lines)):
    lines[l] = lines[l]+',,,'
 
 #Add cats
 if 'cats' in model:
  for col in model['cats']:
   lines[0] = lines[0]+',,,'
   lines[1] = lines[1]+',,'+col+','
   keys = util.get_sorted_keys(model["cats"][col]["uniques"])
   for i in range(len(keys)):
    lines[i+2] = lines[i+2]+',,'+str(keys[i])+','+str(model['cats'][col]['uniques'][keys[i]])
   lines[len(keys)+2] = lines[len(keys)+2] +',,OTHER,'+str(model['cats'][col]['OTHER'])
   for l in range(len(keys)+3, len(lines)):
    lines[l] = lines[l]+',,,'
 
 #Interactions
 if 'catcats' in model:
  for cols in model['catcats']:
   extra = linesify_catcat(model['catcats'][cols])
   lines = lines + [',,'+cols] + extra + ['']
 
 if 'catconts' in model:
  for cols in model['catconts']:
   extra = linesify_catcont(model['catconts'][cols], detail)
   lines = lines + [',,'+cols] + extra + ['']
 
 if 'contconts' in model:
  for cols in model['contconts']:
   extra = linesify_contcont(model['contconts'][cols], detail)
   lines = lines + [',,'+cols] + extra + ['']
 
 #Write to file
 
 f = open(filename, "w")
 for line in lines:
  #print(line)
  f.write(line+'\n')
 f.close()

if __name__ == '__main__':
 exampleModel = {"BASE_VALUE":1700,"conts":{"cont1":[[0.01, 1],[0.02,1.1], [0.03, 1.06]], "cont2":[[37,1.2],[98, 0.9]], "cont3":[[12,1.2],[13, 0.9],[14, 1.1]]}, "cats":{"cat1":{"uniques":{"wstfgl":1.05, "florpalorp":0.92}, "OTHER":1.01},"cat2":{"uniques":{"ruska":1.1, "roma":0.93, 'rita':0.3}, "OTHER":1.9}}, 'catcats':{'cat1 X cat2':{"uniques":{"wstfgl":{"uniques":{"ruska":1.1, "roma":0.93, 'rita':0.3}, "OTHER":1.9}, "florpalorp":{"uniques":{"ruska":1.1, "roma":0.93, 'rita':0.3}, "OTHER":1.9}}, "OTHER":{"uniques":{"ruska":1.1, "roma":0.93, 'rita':0.3}, "OTHER":1.9}}}, 'catconts':{"cat1 X cont1":{"uniques":{"wstfgl":[[0.01, 1],[0.02,1.1], [0.03, 1.06]], "florpalorp":[[0.01, 1],[0.02,1.1], [0.03, 1.06]]}, "OTHER":[[0.01, 1],[0.02,1.1], [0.03, 1.06]]}}, 'contconts':{'cont1 X cont2':[[0.01, [[37,1.2],[98, 0.9]]],[0.02,[[37,1.2],[98, 0.9]]], [0.03, [[37,1.2],[98, 9000]]]]}}
 export_model(exampleModel, 4, 'example.csv')
 
 exampleDf = pd.DataFrame({"cont1":[0.01, 0.012, 0.017, 0.04], "cont2":[40,50,60,70], "cont3":[10,11,12,15], "cat1":["wstfgl","florpalorp","wstfgl","turlingdrome"], "cat2":["ruska","roma","rita","ryla"]})
 uModel = uniquify_model(exampleModel, exampleDf)
 
 print(uModel)
 #print()



import pandas as pd
import numpy as np

from . import actual_modelling
from . import prep
from . import misc
from . import calculus
from . import wraps

#Weighted Poisson proof of concept

df = pd.DataFrame({"cont1":[1,2,3,4,1,2,3,4], "cont2":[1,2,3,4,5,4,3,4], "cat1":['a','a','a','a','b','b','b','a'], "cat2":['c','c','d','d','c','d','e','d'], "exposure":[1,2,1,2,1,2,1,2], "y":[1,2,3,4,5,6,7,8]})

cats = ["cat1", "cat2"]
conts = ["cont1", "cont2"]

model = wraps.prep_model(df, "y", cats, conts)
model = wraps.train_poisson_model(df, 'y', 50, 0.2, model, weightCol="exposure")

#Logistic Proof of Concept 

df = pd.DataFrame({"cont1":[1,2,3,4,1,2,3,4], "cont2":[1,2,3,4,5,4,3,2], "cat1":['a','a','a','a','b','b','b','b'], "cat2":['c','c','d','d','c','d','e','d'], "y":[0,0,0,1,0,0,0,1]})

cats = ["cat1", "cat2"]
conts = ["cont1", "cont2"]

model = wraps.prep_classifier_model(df, "y", cats, conts)
model = wraps.train_classifier_model(df, "y", 50, 0.1, model)

pred = misc.predict(df,model, "Logit")
print(pred)

wraps.interxhunt_classifier_model(df, "y", cats, conts, model, filename="suggestions_logistic")

model = prep.add_catcont_to_model(model, df, 'cat1','cont1')
model = prep.add_catcat_to_model(model, df, 'cat1', 'cat2')
model = prep.add_contcont_to_model(model, df, 'cont1', 'cont2')

wraps.viz_logistic_model(model, "Logistic")

#Saving and Loading

misc.save_model(model)

misc.save_model(model, name="q", timing=False)
m = misc.load_model("models/q.txt")

print(m)
print(model)

#Gamma Proof of Concept

df = pd.DataFrame({"cont1":[1,2,3,4,1,2,3,4], "cont2":[1,2,3,4,5,4,3,4], "cat1":['a','a','a','a','b','b','b','a'], "cat2":['c','c','d','d','c','d','e','d'], "y":[1,2,3,4,5,6,7,8]})

cats = ["cat1", "cat2"]
conts = ["cont1", "cont2"]

model = wraps.prep_model(df,"y",cats,conts)
model = wraps.train_gamma_model(df,"y",50, 0.01, model)
wraps.interxhunt_gamma_model(df, "y", cats, conts, model)

model = prep.add_catcont_to_model(model, df, 'cat1','cont1')
model = prep.add_catcat_to_model(model, df, 'cat1', 'cat2')
model = prep.add_contcont_to_model(model, df, 'cont1', 'cont2')

#Tweedie Proof of Concept

df = pd.DataFrame({"cont1":[1,2,3,4,1,2,3,4], "cont2":[1,2,3,4,5,4,3,4], "cat1":['a','a','a','a','b','b','b','a'], "cat2":['c','c','d','d','c','d','e','d'], "y":[1,2,3,4,5,6,7,8]})

cats = ["cat1", "cat2"]
conts = ["cont1", "cont2"]

model = wraps.prep_model(df,"y",cats,conts)
model = wraps.train_tweedie_model(df,"y",50, 0.01, model)
wraps.interxhunt_tweedie_model(df, "y", cats, conts, model)

model = prep.add_catcont_to_model(model, df, 'cat1','cont1')
model = prep.add_catcat_to_model(model, df, 'cat1', 'cat2')
model = prep.add_contcont_to_model(model, df, 'cont1', 'cont2')

#Multimodel Proof of Concept

df = pd.DataFrame({"cont1":[1,2,3,4,1,2,3,4], "cont2":[1,2,3,4,5,4,3,4], "cat1":['a','a','a','a','b','b','b','a'], "cat2":['c','c','d','d','c','d','e','d'], "y":[1,2,3,4,5,6,7,8]})

cats = ["cat1", "cat2"]
conts = ["cont1", "cont2"]

models = wraps.prep_gamma_models(df, 'y', cats, conts, 2)
models = wraps.train_gamma_models(df, 'y', 50, [0.2,0.3], models)

pred1 = misc.predict(df, models[0])
pred2 = misc.predict(df, models[1])
pred = misc.predict(df, models)

print(pred1, pred2, pred)

wraps.interxhunt_gamma_models(df, "y", cats, conts, models, filename="suggestions_gamma")

models[0] = prep.add_catcont_to_model(models[0], df, 'cat1','cont1')
models[0] = prep.add_catcat_to_model(models[0], df, 'cat1', 'cat2')
models[0] = prep.add_contcont_to_model(models[0], df, 'cont1', 'cont2')

wraps.viz_gamma_models(models, "Gamma")

#Gnormal Proof of Concept

models = wraps.gnormalize_gamma_models(models, df, "y", cats, conts, 11) #I picked 11 because the MPE of the gamma model was a little less than 11%.

models = wraps.train_gnormal_models(df, 'y', 1000, [0.001,0.002,0.001], models)
pred = wraps.predict_from_gnormal(df, models)
predErrPct = wraps.predict_error_from_gnormal(df, models)

print(pred, predErrPct)

wraps.interxhunt_gnormal_models(df,'y',cats,conts,models, filename="suggestions_gnormal")

wraps.viz_gnormal_models(models, "Gnormal")

#Tobit proof of concept

df = pd.read_csv('gnormal.csv')
cdf = df[df['censored']].reset_index()
udf = df[~df['censored']].reset_index()

cats=[]
conts=["x"]

models = wraps.prep_gamma_models(udf, 'y', cats, conts, 1)
models = wraps.train_gamma_models(udf, 'y', 1000, [0.1], models)

print(models)

models = wraps.gnormalize_gamma_models(models, udf, "y", cats, conts, 20)
models = wraps.train_gnormal_models([udf,cdf], 'y', 1000, [0.1,0.005], models)

pred = wraps.predict_from_gnormal(df, models)
predErrPct = wraps.predict_error_from_gnormal(df, models)

df["PREDICTED"]=pred

print(df[["PREDICTED","true_y"]])

print(df["PREDICTED"].mean())
print(df["true_y"].mean())

print(models)

wraps.viz_gnormal_models(models, "Tobit")

#---

#Additive Proof of Concept

df = pd.DataFrame({"cont1":[1,2,3,4,1,2,3,4], "cont2":[1,2,3,4,5,4,3,4], "cat1":['a','a','a','a','b','b','b','a'], "cat2":['c','c','d','d','c','d','e','d'], "y":[1,2,3,4,5,6,7,8]})

cats = ["cat1", "cat2"]
conts = ["cont1", "cont2"]

model = wraps.prep_additive_model(df,'y',cats, conts)

model = wraps.train_normal_model(df,'y', 50, 0.2, model)
pred = misc.predict(df,model)

print(pred)

wraps.interxhunt_normal_model(df, 'y', cats, conts, model, filename="suggestions_additive")

wraps.viz_additive_model(model, "Additive")

#---

#Adjustment Proof of Concept

df["start"] = [1,2,3,4,2.5,3,3.5,8] #We are trying to figure out how best to get from the 'start' column to the 'y' column.

model = wraps.prep_adjustment_model(df,'y', 'start', cats, conts)

print(model)

model = wraps.train_adjustment_model(df, 'y', 'start', 50, 0.2, model)

print(model)

pred = misc.predict(df,model)

print(pred)

wraps.interxhunt_adjustment_model(df, 'y', cats, conts, model, filename="suggestions_adj")

wraps.viz_adjustment_model(model, "Adjustment")

#Gnormalized Adjustment Proof of Concept

models = wraps.gnormalize_gamma_models([model], df, "y", cats, conts, 4)

print(pred, predErrPct)

models = wraps.train_gnormal_models(df, 'y', 2000, [0.0001,0.0001], models, staticFeats=["start"], prints="verbose")

print(models)
pred = wraps.predict_from_gnormal(df, models)
predErrPct = wraps.predict_error_from_gnormal(df, models)

print(pred, predErrPct)

wraps.interxhunt_gnormal_models(df,'y',cats,conts,models, filename="suggestions_gnormadj")

wraps.viz_gnormal_models(models, "Gnormallized Adjustment")

#---

#Penalization Proof of Concept

df = pd.DataFrame({"cont1":[1,4,3,2,6,7,5,9,8], "cont2":[1,2,3,1,2,3,1,2,3], "cont3":[1,2,3,4,5,4,3,2,1]})
df["y"]= df["cont1"] + df["cont2"]*0.04 + df["cont3"]*0.01

cats=[]
conts=["cont1","cont2", "cont3"]

model = wraps.prep_additive_model(df,'y',cats, conts)

#The intended workflow is as follows:
model = wraps.train_normal_model(df,'y', 50, 0.2, model, pen=0.8)# First, train with high penalization . . .
model = misc.de_feat(model, 0) # . . . then, purge all features that LASSO pulled to the default . . . 
model = wraps.train_normal_model(df,'y', 50, 0.2, model, pen=0) # . . . and re-train on remaining features with lower (or zero!) penalization.

#(NOTE #1: for a multiplicative model the middle line would be "misc.de_feat(model, 1)", as the default value there is 1.)
#(NOTE #2: de_feat doesn't affect interactions because if you're using penalization to remove interaction effects then you have made poor life choices.)

print(model)

print(misc.what_conts(model)) 
print(misc.what_cats(model))
print(misc.how_many_conts(model))
print(misc.how_many_cats(model))
#"Oh hey, looks like only one feature survived!" (You can vary penalization until you have roughly the number of features you want, of the kinds you want.)

pred = misc.predict(df,model)
print(pred)

wraps.viz_additive_model(model, "Penalized")

#---

#Cratio Proof of Concept

df = pd.DataFrame({"cont1":[1,2,3,4,5,6,7,8,9], "cont2":[1,2,3,1,2,3,1,2,3], "cont3":[1,2,3,4,5,4,3,2,1]})
df["y"]= ((df["cont3"]>3)|((df["cont2"]>2)&(df["cont1"]>4))).astype(int)

cats=[]
conts=["cont1","cont2", "cont3"]

models = wraps.prep_cratio_models(df, 'y', cats, conts, 1)

models = wraps.train_cratio_models(df, 'y', 100, [0.2], models)

print(models)

pred = misc.predict(df,models, calculus.Cratio_mlink)

print(pred)

predRat = misc.predict(df,models)

print(predRat)

print(df["y"])


wraps.viz_cratio_models(models, "Cratio")

#---

#Flink Proof of Concept

df = pd.DataFrame({"cont1":[0,1,0,1],"cat1":[0,0,1,1],"y":[0,1,1,4]})
model = wraps.prep_additive_model(df, "y", ["cat1"], ["cont1"])
model = wraps.train_normal_model(df, "y", 100, 0.1, model)
print(model)
print(misc.predict(df, model))
model = wraps.flinkify_additive_model(df,model)
model = wraps.train_normal_model(df, "y", 100, 0.1, model)
print(model)
print(misc.predict(df, model))
wraps.viz_additive_model(model, "flink")

#Flink also works for multiples, in both senses of the word!



import pandas as pd
import numpy as np
import math

def get_weighted_gini(df, predCol, actCol, weightCol, tiebreaks="pessimistic"):
 df['meanAct'] = sum(df[actCol]*df[weightCol])/sum(df[weightCol])
 
 if tiebreaks=="pessimistic":
  df = df.sort_values([predCol, actCol], ascending=[False,True])
 elif tiebreaks=="optimistic":
  df = df.sort_values([predCol, actCol], ascending=[False,False])
 else:
  df = df.sort_values(predCol, ascending=False)
 
 df['diff'] = df[actCol]-df['meanAct']
 df['gap'] = (df['diff']*df[weightCol]).cumsum()
 df['base'] = (df['meanAct']*df[weightCol]).cumsum()
 return sum(df['gap']*df[weightCol])/sum(df['base']*df[weightCol])

def get_weighted_Xiles(df, predCol, actCol, weightCol, X=10, tiebreaks="pessimistic"):
 if tiebreaks=="pessimistic":
  df = df.sort_values([predCol, actCol], ascending=[True,False])
 elif tiebreaks=="optimistic":
  df = df.sort_values([predCol, actCol], ascending=[True,True])
 else:
  df = df.sort_values(predCol, ascending=True)
 
 totWeight = sum(df[weightCol])
 df['cumWeight'] = df[weightCol].cumsum()
 p=[]
 a=[]
 for i in range(X):
  subDf = df[df['cumWeight']>(i*totWeight/X)].reset_index(drop=True)
  subDf = subDf[subDf['cumWeight']<=((i+1)*totWeight/X)]
  p.append(sum(subDf[predCol]*subDf[weightCol])/sum(subDf[weightCol]))
  a.append(sum(subDf[actCol]*subDf[weightCol])/sum(subDf[weightCol]))
 return p, a

def get_gini(df, predCol, actCol, tiebreaks="pessimistic"):
 df["w8"] = 1
 return get_weighted_gini(df, predCol, actCol, 'w8', tiebreaks)

def get_Xiles(df, predCol, actCol, X=10, tiebreaks="pessimistic"):
 df["w8"] = 1
 return get_weighted_Xiles(df, predCol, actCol, 'w8', X, tiebreaks)

#---

def get_weighted_MAE(df, predCol, actCol, weightCol):
 return ((df[predCol]-df[actCol]).abs()*df[weightCol]).sum()/sum(df[weightCol])

def get_weighted_RMSE(df, predCol, actCol, weightCol):
 return math.sqrt(((df[predCol]-df[actCol])*(df[predCol]-df[actCol])*df[weightCol]).sum()/sum(df[weightCol]))

def get_weighted_MPE(df, predCol, actCol, weightCol): #Yes, I know the standard definition uses Actual as the denominator. The standard definition is wrong.
 PE = 100*(df[predCol]-df[actCol])/df[predCol]
 return (PE.abs()*df[weightCol]).sum()/sum(df[weightCol])

def get_weighted_RMSPE(df, predCol, actCol, weightCol): 
 PE = 100*(df[predCol]-df[actCol])/df[predCol]
 return math.sqrt((PE*PE*df[weightCol]).sum()/sum(df[weightCol]))

def get_MAE(df, predCol, actCol):
 df["w8"] = 1
 return get_weighted_MAE(df, predCol, actCol, "w8")

def get_RMSE(df, predCol, actCol):
 df["w8"] = 1
 return get_weighted_RMSE(df, predCol, actCol, "w8")

def get_MPE(df, predCol, actCol):
 df["w8"] = 1
 return get_weighted_MPE(df, predCol, actCol, "w8")

def get_RMSPE(df, predCol, actCol):
 df["w8"] = 1
 return get_weighted_RMSPE(df, predCol, actCol, "w8")
 

if __name__ == '__main__':
 df = pd.DataFrame({"P":[1,2,3,4,5,7,6,8,9],"A":[1,2,3,4,5,6,8,7,9],"W":[1,2,3,4,5,6,7,8,9]})
 print(get_weighted_gini(df, 'P','A','W'))
 print(get_weighted_gini(df, 'A','A','W'))
 print(get_weighted_gini(df, 'P','A','W')/get_weighted_gini(df, 'A','A','W'))
 
 print(get_weighted_Xiles(df, 'P','A','W', 3))
 
 
 df = pd.DataFrame({"P":[1,3,3,3,5],"A":[1,2,4,3,5]})
 print(get_gini(df,'P','A', "pessimistic"))
 print(get_gini(df,'P','A', "optimistic"))
 print(get_gini(df,'P','A', "whatever"))
 
 print(get_Xiles(df,'P','A', 5, "pessimistic"))
 print(get_Xiles(df,'P','A', 5, "optimistic"))
 print(get_Xiles(df,'P','A', 5, "whatever"))
 



import durkon.actual_modelling
import durkon.calculus
import durkon.export
import durkon.metrics
import durkon.misc
import durkon.pena
import durkon.prep
import durkon.rele
import durkon.util
import durkon.viz
import durkon.wraps


import pandas as pd
import numpy as np
import math
import copy
import time

from . import util
from . import misc
from . import calculus
from . import pena
from . import rele
from . import prep

#pens, lrs, staticFeatLists, specificPenses and models are per model
#relelistlists are listed per df, then per model
#link, linkgrad are per pred; linkgrad is then per model
#lossgrad is per df, then per pred

def train_models(inputDfs, targets, nrounds, lrs, startingModels, weightCol=None, staticFeats = [], lras=None, lossgrads=[[calculus.Gauss_grad]], links=[calculus.Unity_link], linkgrads=[[calculus.Unity_link_grad]], pens=None, minRela=None, prints="verbose", record=False):
 
 history=[]
 
 inputDfs = [df.reset_index() for df in inputDfs]
 
 if type(targets)!=type([1,2,3]):
  targets=[targets]
 
 if lras==None:
  lras = [calculus.default_LRA]*len(startingModels)
 
 models = copy.deepcopy(startingModels)
 
 if prints!="silent":
  print("initial weights and relevances setup")
 
 contReleDictListList = []
 catReleDictListList = []
 totReleDictListList = []
 
 contWReleDictListList = []
 catWReleDictListList = []
 totWReleDictListList = []
 
 interReleDictListList = []
 totInterReleDictListList = []
 
 interWReleDictListList = []
 totInterWReleDictListList = []
 
 ws=[]
 
 for inputDf in inputDfs:
  if weightCol==None:
   weight = np.ones(len(inputDf))
  else:
   weight = inputDf[weightCol]
  w = np.array(np.transpose(np.matrix(weight)))
  
  ws.append(w)
  
  contReleDictList = []
  catReleDictList = []
  totReleDictList = []
  
  contWReleDictList = []
  catWReleDictList = []
  totWReleDictList = []
  
  interReleDictList = []
  totInterReleDictList = []
  
  interWReleDictList = []
  totInterWReleDictList = []
  
  for model in models:
   
   cord = rele.produce_cont_relevances_dict(inputDf,model) #d(feat)/d(pt)
   card = rele.produce_cat_relevances_dict(inputDf,model) #d(feat)/d(pt)
   tord = rele.produce_total_relevances_dict(cord, card)
   
   cowrd = rele.produce_wReleDict(cord, w) #d(feat)/d(pt), adjusted for weighting
   cawrd = rele.produce_wReleDict(card, w) #d(feat)/d(pt), adjusted for weighting
   towrd = rele.produce_total_relevances_dict(cowrd, cawrd)
   
   #Interactions . . .
   
   ird = rele.produce_interxn_relevances_dict(inputDf, model) #d(feat)/d(pt)
   tird = rele.produce_total_irelevances_dict(ird)
   
   wird = rele.produce_wReleDict(ird, w) #d(feat)/d(pt), adjusted for weighting
   twird = rele.produce_total_irelevances_dict(wird)
   
   contReleDictList.append(copy.deepcopy(cord))
   catReleDictList.append(copy.deepcopy(card))
   totReleDictList.append(copy.deepcopy(tord))
   
   contWReleDictList.append(copy.deepcopy(cowrd))
   catWReleDictList.append(copy.deepcopy(cawrd))
   totWReleDictList.append(copy.deepcopy(towrd))
   
   interReleDictList.append(copy.deepcopy(ird))
   totInterReleDictList.append(copy.deepcopy(tird))
   
   interWReleDictList.append(copy.deepcopy(wird))
   totInterWReleDictList.append(copy.deepcopy(twird))
  
  contReleDictListList.append(copy.deepcopy(contReleDictList))
  catReleDictListList.append(copy.deepcopy(catReleDictList))
  totReleDictListList.append(copy.deepcopy(totReleDictList))
  
  contWReleDictListList.append(copy.deepcopy(contWReleDictList))
  catWReleDictListList.append(copy.deepcopy(catWReleDictList))
  totWReleDictListList.append(copy.deepcopy(totWReleDictList))
  
  interReleDictListList.append(copy.deepcopy(interReleDictList))
  totInterReleDictListList.append(copy.deepcopy(totInterReleDictList))
  
  interWReleDictListList.append(copy.deepcopy(interWReleDictList))
  totInterWReleDictListList.append(copy.deepcopy(totInterWReleDictList))
  
  
 for i in range(nrounds):
  
  if prints!="silent":
   print("epoch: "+str(i+1)+"/"+str(nrounds))
  if prints=="verbose":
   for model in models:
    misc.explain(model)
  
  oModels=copy.deepcopy(models) #Duplicate so effects for all dfs are applied simultaneously
  
  for d in range(len(inputDfs)):
   
   inputDf = inputDfs[d]
   
   if (prints=="verbose") and (len(inputDfs)>1):
    print("adjusting to df #"+ str(d))
   
   if prints=="verbose":
    print("initial pred and effect-gathering")
   
   contEffectsList = []
   catEffectsList = []
   interxEffectsList = []
   
   preFlinkCombs=[]
   combs = []
   
   for m in range(len(models)):
    
    oModel = oModels[m]
    
    contEffects = misc.get_effects_of_cont_cols_from_relevance_dict(contReleDictListList[d][m],oModel)
    catEffects = misc.get_effects_of_cat_cols_from_relevance_dict(catReleDictListList[d][m],oModel)
    interxEffects = misc.get_effects_of_interxns_from_relevance_dict(interReleDictListList[d][m],oModel)
    
    contEffectsList.append(contEffects)
    catEffectsList.append(catEffects)
    interxEffectsList.append(interxEffects)
    
    if oModel['featcomb']=="addl":
     comb = misc.comb_from_effects_addl(oModel["BASE_VALUE"], len(inputDf), contEffects, catEffects, interxEffects)
    else:
     comb = misc.comb_from_effects_mult(oModel["BASE_VALUE"], len(inputDf), contEffects, catEffects, interxEffects)
    
    if "flink" in oModel:
     flinkEffect = misc.get_effect_of_this_cont_col(comb, oModel["flink"])
     preFlinkComb = comb
     if oModel['featcomb']=="addl":
      comb = comb+flinkEffect
     else:
      comb = comb*flinkEffect
     preFlinkCombs.append(preFlinkComb)
    else:
     preFlinkCombs.append({})
    
    combs.append(comb)
   
   #linkgradients = [[misc.lambdapply(linkgrad, combs) for linkgrad in linkgradset] for linkgradset in linkgrads]
   #preds = [misc.lambdapply(link, combs) for link in links]
   #lossgradients = [misc.lambdapply(lossgrad, preds+[df[target]]) for lossgrad in lossgrads[d]]
   #
   #print("linkg", linkgradients)
   #print('p', preds)
   #print("lossg", lossgradients)
   
   
   linkgradients = [[linkgrad(*combs) for linkgrad in linkgradset] for linkgradset in linkgrads]
   preds = [link(*combs) for link in links]
   tCols = [inputDf[target] for target in targets]
   lossgradients = [lossgrad(*preds,*tCols) for lossgrad in lossgrads[d]]
   
   #print(lossgradients)
   
   for p in range(len(preds)):
    
    for m in range(len(models)):
     
     model = models[m]
     
     lra = lras[m](*combs)
     
     cowrd = contWReleDictListList[d][m]
     cawrd = catWReleDictListList[d][m]
     wird = interWReleDictListList[d][m]
     
     linkgradient = linkgradients[p][m]
     lossgradient = lossgradients[p]
     
     if type(linkgradient)==type(pd.Series([])):
      
      if "flink" in model:
       if prints=="verbose":
        print("adjust flink")
       
       flinkRele = rele.produce_cont_relevances(preFlinkCombs[m], model["flink"])#Has to be reproduced every turn because the dist changes over time (it better!)
       
       if model["featcomb"]=="addl":
        finalGradients = np.matmul(np.array(lossgradient*linkgradient),ws[d]*flinkRele)#d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
       else:
        finalGradients = np.matmul(np.array(lossgradient*linkgradient*preFlinkCombs[m]),ws[d]*flinkRele) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
       
       
       totReles = rele.sum_and_listify_matrix(ws[d]*flinkRele)
       for k in range(len(model['flink'])):
        totRele = totReles[k]
        if totRele!=0:
         model["flink"][k][1] -= finalGradients[k]*lrs[m]/totRele
      
      if "conts" in model:
       if prints=="verbose":
        print("adjust conts")
       
       for col in [c for c in model['conts'] if c not in staticFeats]:
        
        if model['featcomb']=="addl":
         finalGradients = np.matmul(np.array(lossgradient*linkgradient),cowrd[col]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
        else:
         effectOfCol = contEffectsList[m][col]
         ceoc = combs[m]/effectOfCol #d(comb)/d(feat)
         finalGradients = np.matmul(np.array(lossgradient*linkgradient*ceoc),cowrd[col]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
        
        for k in range(len(model['conts'][col])):
         totRele = sum([totWReleDictListList[D][m]["conts"][col][k] for D in range(len(inputDfs))])
         if totRele>0:
          model["conts"][col][k][1] -= finalGradients[k]*lrs[m]*lra/totRele
      
      if "cats" in model:
       if prints=="verbose":
        print("adjust cats")
       
       for col in [c for c in model['cats'] if c not in staticFeats]:
        
        if model['featcomb']=="addl":
         finalGradients = np.matmul(np.array(lossgradient*linkgradient),cawrd[col]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
        else:
         effectOfCol = catEffectsList[m][col]
         ceoc = combs[m]/effectOfCol #d(comb)/d(feat)
         finalGradients = np.matmul(np.array(lossgradient*linkgradient*ceoc),cawrd[col]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
        
        skeys = util.get_sorted_keys(model['cats'][col]["uniques"])
        
        #all the uniques . . .
        for k in range(len(skeys)):
         totRele = sum([totWReleDictListList[D][m]["cats"][col][k] for D in range(len(inputDfs))])
         if totRele>0:
          model["cats"][col]["uniques"][skeys[k]] -= finalGradients[k]*lrs[m]*lra/totRele
         
        # . . . and "OTHER"
        totRele = towrd["cats"][col][-1]
        if totRele>0:
         model["cats"][col]["OTHER"] -= finalGradients[-1]*lrs[m]*lra/totRele
        
      if "catcats" in model:
       if prints=="verbose":
        print('adjust catcats')
       
       for cols in [c for c in model['catcats'] if c not in staticFeats]:
        
        if model['featcomb']=="addl":
         finalGradients = np.matmul(np.array(lossgradient*linkgradient),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
        else:
         effectOfCols = interxEffectsList[m][cols]
         ceoc = combs[m]/effectOfCols #d(comb)/d(feat)
         finalGradients = np.matmul(np.array(lossgradient*linkgradient*ceoc),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
        
        skeys1 = util.get_sorted_keys(model['catcats'][cols]["uniques"])
        skeys2 = util.get_sorted_keys(model['catcats'][cols]["OTHER"]["uniques"])
        
        for i in range(len(skeys1)):
         for j in range(len(skeys2)):
          totRele = sum([totInterWReleDictListList[D][m][cols][i*(len(skeys2)+1)+j] for D in range(len(inputDfs))])
          if totRele>0:
           model["catcats"][cols]["uniques"][skeys1[i]]['uniques'][skeys2[j]] -= finalGradients[i*(len(skeys2)+1)+j]*lrs[m]*lra/totRele
         totRele = sum([totInterWReleDictListList[D][m][cols][i*(len(skeys2)+1)+len(skeys2)] for D in range(len(inputDfs))])
         if totRele>0:
          model['catcats'][cols]["uniques"][skeys1[i]]['OTHER'] -= finalGradients[i*(len(skeys2)+1)+len(skeys2)]*lrs[m]*lra/totRele
        
        for j in range(len(skeys2)):
         totRele = sum([totInterWReleDictListList[D][m][cols][len(skeys1)*(len(skeys2)+1)+j] for D in range(len(inputDfs))])
         if totRele>0:
          model["catcats"][cols]['OTHER']['uniques'][skeys2[j]] -= finalGradients[len(skeys1)*(len(skeys2)+1)+j]*lrs[m]*lra/totRele
        
        totRele = sum([ totInterWReleDictListList[D][m][cols][-1] for D in range(len(inputDfs))])
        if totRele>0:
         model['catcats'][cols]['OTHER']['OTHER'] -= finalGradients[-1]*lrs[m]*lra/totRele
      
      if "catconts" in model:
       if prints=="verbose":
        print('adjust catconts')
       
       for cols in [c for c in model['catconts'] if c not in staticFeats]:
        
        if model['featcomb']=="addl":
         finalGradients = np.matmul(np.array(lossgradient*linkgradient),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
        else:
         effectOfCols = interxEffectsList[m][cols]
         ceoc = combs[m]/effectOfCols #d(comb)/d(feat)
         finalGradients = np.matmul(np.array(lossgradient*linkgradient*ceoc),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
        
        skeys = util.get_sorted_keys(model['catconts'][cols]["uniques"])
        
        for i in range(len(skeys)):
         for j in range(len(model['catconts'][cols]["OTHER"])):
          totRele = sum([totInterWReleDictListList[D][m][cols][i*(len(model['catconts'][cols]["OTHER"])+1)+j] for D in range(len(inputDfs))])
          if totRele>0:
           model['catconts'][cols]['uniques'][skeys[i]][j][1] -= finalGradients[i*(len(model['catconts'][cols]["OTHER"])+1)+j]*lrs[m]*lra/totRele
        
        for j in range(len(model['catconts'][cols]["OTHER"])):
         totRele = sum([totInterWReleDictListList[D][m][cols][len(skeys)*(len(model['catconts'][cols]["OTHER"])+1)+j] for D in range(len(inputDfs))])
         if totRele>0:
          model['catconts'][cols]['OTHER'][j][1] -= finalGradients[len(skeys)*(len(model['catconts'][cols]["OTHER"])+1)+j]*lrs[m]*lra/totRele
      
      if "contconts" in model:
       if prints=="verbose":
        print('adjust contconts')
       
       for cols in [c for c in model['contconts'] if c not in staticFeats]:
        
        if model['featcomb']=="addl":
         finalGradients = np.matmul(np.array(lossgradient*linkgradient),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
        else:
         effectOfCols = interxEffectsList[m][cols]
         ceoc = combs[m]/effectOfCols #d(comb)/d(feat)
         finalGradients = np.matmul(np.array(lossgradient*linkgradient*ceoc),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
        
        for i in range(len(model['contconts'][cols])):
         for j in range(len(model['contconts'][cols][0][1])):
          totRele = sum([totInterWReleDictListList[D][m][cols][i*(len(model['contconts'][cols][0][1])+1)+j] for D in range(len(inputDfs))])
          if totRele>0:
           model['contconts'][cols][i][1][j][1] -= finalGradients[i*(len(model['contconts'][cols][0][1])+1)+j]*lrs[m]*lra/totRele
  
  #Penalize!
  if pens!=None:
   if prints=="verbose":
    print("penalties")
   for m in range(len(models)):
    if models[m]['featcomb']=="addl":
     models[m] = pena.penalize_model(models[m], pens[m]*lrs[m], 0)
    else:
     models[m] = pena.penalize_model(models[m], pens[m]*lrs[m], 1)
  
  if minRela!=None:
   models = [misc.enforce_min_rela(model, minRela) for model in models]
  
  history.append(models)
 
 if record: 
  return models, history
 return models

def train_model(inputDf, target, nrounds, lr, startingModel, weightCol=None, staticFeats = [], pen=0, specificPens={}, lossgrad=calculus.Poisson_grad, link=calculus.Unity_link, linkgrad=calculus.Unity_link_grad, minRela=None, prints="verbose", record=False):
 
 history=[]
 
 inputDf = inputDf.reset_index()
 
 model = copy.deepcopy(startingModel)
 
 if "featcomb" not in model:
  model["featcomb"]="mult"
 
 if prints!="silent":
  print("initial weights and relevances setup")
 
 if weightCol==None:
  weight = np.ones(len(inputDf))
 else:
  weight = inputDf[weightCol]
 w = np.array(np.transpose(np.matrix(weight)))
 
 cord = rele.produce_cont_relevances_dict(inputDf,model) #d(feat)/d(pt)
 card = rele.produce_cat_relevances_dict(inputDf,model) #d(feat)/d(pt)
 tord = rele.produce_total_relevances_dict(cord, card)
 
 cowrd = rele.produce_wReleDict(cord, w) #d(feat)/d(pt), adjusted for weighting
 cawrd = rele.produce_wReleDict(card, w) #d(feat)/d(pt), adjusted for weighting
 towrd = rele.produce_total_relevances_dict(cowrd, cawrd)
 
 #Interactions . . .
 
 ird = rele.produce_interxn_relevances_dict(inputDf, model) #d(feat)/d(pt)
 tird = rele.produce_total_irelevances_dict(ird)
 
 wird = rele.produce_wReleDict(ird, w) #d(feat)/d(pt), adjusted for weighting
 twird = rele.produce_total_irelevances_dict(wird)
 
 for i in range(nrounds):
  
  if prints!="silent":
   print("epoch: "+str(i+1)+"/"+str(nrounds))
  if prints=="verbose":
   misc.explain(model)
  
  if prints=="verbose":
   print("initial pred and effect-gathering")
  
  contEffects = misc.get_effects_of_cont_cols_from_relevance_dict(cord,model)
  catEffects = misc.get_effects_of_cat_cols_from_relevance_dict(card,model)
  interxEffects = misc.get_effects_of_interxns_from_relevance_dict(ird,model)
  
  if model['featcomb']=="addl":
   comb = misc.comb_from_effects_addl(model["BASE_VALUE"], len(inputDf), contEffects, catEffects, interxEffects)
  else:
   comb = misc.comb_from_effects_mult(model["BASE_VALUE"], len(inputDf), contEffects, catEffects, interxEffects)
  
  if "flink" in model:
   flinkEffect = misc.get_effect_of_this_cont_col(comb, model["flink"])
   preFlinkComb = comb
   if model['featcomb']=="addl":
    comb = comb+flinkEffect
   else:
    comb = comb*flinkEffect
  
  linkgradient = linkgrad(comb) #d(pred)/d(comb)
  
  pred = comb.apply(link)
  lossgradient = lossgrad(pred, np.array(inputDf[target])) #d(Loss)/d(pred)
  
  #Adjust adjustable model components one-by-one . . .
  
  if "flink" in model:
   if prints=="verbose":
    print("adjust flink")
   
   flinkRele = rele.produce_cont_relevances(preFlinkComb, model["flink"])#Has to be reproduced every turn because the dist changes over time (it better!)
   
   if model["featcomb"]=="addl":
    finalGradients = np.matmul(np.array(lossgradient*linkgradient),w*flinkRele)#d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
   else:
    finalGradients = np.matmul(np.array(lossgradient*linkgradient*preFlinkComb),w*flinkRele) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
   
   
   totReles = rele.sum_and_listify_matrix(w*flinkRele)
   for k in range(len(model['flink'])):
    totRele = totReles[k]
    if totRele!=0:
     model["flink"][k][1] -= finalGradients[k]*lr/totRele
  
  if "conts" in model:
   if prints=="verbose":
    print("adjust conts")
   
   for col in [c for c in model['conts'] if c not in staticFeats]:
    
    if model['featcomb']=="addl":
     finalGradients = np.matmul(np.array(lossgradient*linkgradient),cowrd[col]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
    else:
     effectOfCol = contEffects[col]
     ceoc = comb/effectOfCol #d(comb)/d(feat)
     finalGradients = np.matmul(np.array(lossgradient*linkgradient*ceoc),cowrd[col]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
    
    for k in range(len(model['conts'][col])):
     totRele = towrd["conts"][col][k]
     if totRele>0:
      model["conts"][col][k][1] -= finalGradients[k]*lr/totRele
  
  if "cats" in model:
   if prints=="verbose":
    print("adjust cats")
   
   for col in [c for c in model['cats'] if c not in staticFeats]:
    
    if model['featcomb']=="addl":
     finalGradients = np.matmul(np.array(lossgradient*linkgradient),cawrd[col]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
    else:
     effectOfCol = catEffects[col]
     ceoc = comb/effectOfCol #d(comb)/d(feat)
     finalGradients = np.matmul(np.array(lossgradient*linkgradient*ceoc),cawrd[col]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
    
    skeys = util.get_sorted_keys(model['cats'][col]["uniques"])
    
    #all the uniques . . .
    for k in range(len(skeys)):
     totRele = towrd["cats"][col][k]
     if totRele>0:
      model["cats"][col]["uniques"][skeys[k]] -= finalGradients[k]*lr/totRele
     
    # . . . and "OTHER"
    totRele = towrd["cats"][col][-1]
    if totRele>0:
     model["cats"][col]["OTHER"] -= finalGradients[-1]*lr/totRele
    
  if "catcats" in model:
   if prints=="verbose":
    print('adjust catcats')
   
   for cols in [c for c in model['catcats'] if c not in staticFeats]:
    
    if model['featcomb']=="addl":
     finalGradients = np.matmul(np.array(lossgradient*linkgradient),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
    else:
     effectOfCols = interxEffects[cols]
     ceoc = comb/effectOfCols #d(comb)/d(feat)
     finalGradients = np.matmul(np.array(lossgradient*linkgradient*ceoc),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
    
    skeys1 = util.get_sorted_keys(model['catcats'][cols]["uniques"])
    skeys2 = util.get_sorted_keys(model['catcats'][cols]["OTHER"]["uniques"])
    
    for i in range(len(skeys1)):
     for j in range(len(skeys2)):
      totRele = twird[cols][i*(len(skeys2)+1)+j]
      if totRele>0:
       model["catcats"][cols]["uniques"][skeys1[i]]['uniques'][skeys2[j]] -= finalGradients[i*(len(skeys2)+1)+j]*lr/totRele
     totRele = twird[cols][i*(len(skeys2)+1)+len(skeys2)]
     if totRele>0:
      model['catcats'][cols]["uniques"][skeys1[i]]['OTHER'] -= finalGradients[i*(len(skeys2)+1)+len(skeys2)]*lr/totRele
    
    for j in range(len(skeys2)):
     totRele = twird[cols][len(skeys1)*(len(skeys2)+1)+j]
     if totRele>0:
      model["catcats"][cols]['OTHER']['uniques'][skeys2[j]] -= finalGradients[len(skeys1)*(len(skeys2)+1)+j]*lr/totRele
     
    totRele = twird[cols][-1]
    if totRele>0:
     model['catcats'][cols]['OTHER']['OTHER'] -= finalGradients[-1]*lr/totRele
  
  if "catconts" in model:
   if prints=="verbose":
    print('adjust catconts')
  
   for cols in [c for c in model['catconts'] if c not in staticFeats]:
    
    if model['featcomb']=="addl":
     finalGradients = np.matmul(np.array(lossgradient*linkgradient),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
    else:
     effectOfCols = interxEffects[cols]
     ceoc = comb/effectOfCols #d(comb)/d(feat)
     finalGradients = np.matmul(np.array(lossgradient*linkgradient*ceoc),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
    
    skeys = util.get_sorted_keys(model['catconts'][cols]["uniques"])
    
    for i in range(len(skeys)):
     for j in range(len(model['catconts'][cols]["OTHER"])):
      totRele = twird[cols][i*(len(model['catconts'][cols]["OTHER"])+1)+j]
      if totRele>0:
       model['catconts'][cols]['uniques'][skeys[i]][j][1] -= finalGradients[i*(len(model['catconts'][cols]["OTHER"])+1)+j]*lr/totRele
    
    for j in range(len(model['catconts'][cols]["OTHER"])):
     totRele = twird[cols][len(skeys)*(len(model['catconts'][cols]["OTHER"])+1)+j]
     if totRele>0:
      model['catconts'][cols]['OTHER'][j][1] -= finalGradients[len(skeys)*(len(model['catconts'][cols]["OTHER"])+1)+j]*lr/totRele
  
  if "contconts" in model:
   if prints=="verbose":
    print('adjust contconts')
   
   for cols in [c for c in model['contconts'] if c not in staticFeats]:
    
    if model['featcomb']=="addl":
     finalGradients = np.matmul(np.array(lossgradient*linkgradient),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb) * d(comb)/d(pt)
    else:
     effectOfCols = interxEffects[cols]
     ceoc = comb/effectOfCols #d(comb)/d(feat)
     finalGradients = np.matmul(np.array(lossgradient*linkgradient*ceoc),wird[cols]) #d(Loss)/d(pt) = d(Loss)/d(pred) * d(pred)/d(comb)* d(comb)/d(feat) * d(feat)/d(pt)
    
    for i in range(len(model['contconts'][cols])):
     for j in range(len(model['contconts'][cols][0][1])):
      totRele = twird[cols][i*(len(model['contconts'][cols][0][1])+1)+j]
      if totRele>0:
       model['contconts'][cols][i][1][j][1] -= finalGradients[i*(len(model['contconts'][cols][0][1])+1)+j]*lr/totRele
  
  #Penalize!
  if pen>0:
   if prints=="verbose":
    print("penalties")
   if model['featcomb']=="addl":
    model = pena.penalize_model(model, pen*lr, 0, specificPens)
   else:
    model = pena.penalize_model(model, pen*lr, 1, specificPens)
   
   if minRela!=None:
    model = misc.enforce_min_rela(model, minRela)
 
  history.append(model)
 
 if record:
  return model, history
 return model


if __name__ == '__main__':
 df = pd.DataFrame({"cont1":[0,1,0,1],"cat1":[0,0,1,1],"y":[0,1,1,4]})
 model = {"BASE_VALUE":1.0,"conts":{"cont1":[[0,0],[1,1]]}, "cats":{"cat1":{"uniques":{0:0, 1:0},"OTHER":0}},'featcomb':'addl'}
 model = train_model(df, "y", 100, 0.1, model, lossgrad=calculus.Gauss_grad)
 print(model)
 print(misc.predict(df, model))
 df["preds"] = misc.predict(df, model)
 model["flink"] = prep.get_cont_feat(df, "preds",3, 0.001, 0)
 model = train_model(df, "y", 100, 0.1, model, lossgrad=calculus.Gauss_grad)
 print(model)
 print(misc.predict(df, model))





if False: #__name__ == '__main__':
 df = pd.DataFrame({"x":[1,2,3],"y":[2,3,4]})
 models = [{"BASE_VALUE":1.0,"conts":{"x":[[0,1],[4,1]]}, "cats":{},'featcomb':'mult'}]
 newModels = train_models([df], "y",100, [0.02], models)
 for newModel in newModels:
  misc.explain(newModel)
 
 df = pd.DataFrame({"x1":[1,2,3,4],"x2":[1,2,3,2],"y":[1+1,2+2,3+3,4+2]})
 models = [{"BASE_VALUE":2,"conts":{"x1":[[1,1],[4,1]]}, "cats":{},'featcomb':'mult'}, {"BASE_VALUE":1,"conts":{"x2":[[1,1],[4,1]]}, "cats":{},'featcomb':'mult'}]
 newModels = train_models([df], "y",200, [0.1, 0.1], models, links=[calculus.Add_mlink], linkgrads=[[calculus.Add_mlink_grad, calculus.Add_mlink_grad]])
 print(newModels)
 for newModel in newModels:
  misc.explain(newModel)
 
 print(misc.predict_models(df, newModels, calculus.Add_mlink)) #It looks like it fails but it doesn't, put whatever base values or varying lrs you want and it'll appear to crash while predicting perfectly.
 
 df = pd.DataFrame({"x1":[1,2,3,4,1,2,3,4],"x2":[1,1,2,2,3,3,4,4],"y":[max(1,1),max(2,1),max(3,2),max(4,2),max(1,3),max(2,3),max(3,4),max(4,4)]})
 models = [{"BASE_VALUE":1,"conts":{"x1":[[1,1],[4,1]]}, "cats":{},'featcomb':'mult'}, {"BASE_VALUE":1,"conts":{"x2":[[1,1],[4,1]]}, "cats":{},'featcomb':'mult'}]
 newModels = train_models([df], "y",400, [0.1, 0.12], models, links=[calculus.Max_mlink_2], linkgrads=[[calculus.Max_mlink_grad_2_A, calculus.Max_mlink_grad_2_B]])
 print(newModels)
 for newModel in newModels:
  misc.explain(newModel)
 
 df = pd.DataFrame({"x1":[1,2,3,4,1,2,3,4,1,2,3,4,1,2,3,4],"x2":[1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,4],"y":[1*1,2*1,3*1,4*1,1*2,2*2,3*2,4*2,1*3,2*3,3*3,4*3,1*4,2*4,3*4,4*4]})
 models = [{"BASE_VALUE":-3,"conts":{"x1":[[1,1],[4,1]]}, "cats":{},'featcomb':'addl'}, {"BASE_VALUE":-2,"conts":{"x2":[[1,1],[4,1]]}, "cats":{},'featcomb':'addl'}]
 newModels = train_models([df], "y",200, [0.01, 0.02], models, links=[calculus.Mult_mlink_2], linkgrads=[[calculus.Mult_mlink_grad_2_A, calculus.Mult_mlink_grad_2_B]]) 
 print(newModels)
 for newModel in newModels:
  misc.explain(newModel)
 
 print(misc.predict_models(df, newModels, calculus.Mult_mlink_2)) #It looks like it fails but it doesn't, put whatever base values or varying lrs you want and it'll appear to crash while predicting perfectly.
 
 df = pd.read_csv('gnormal.csv')
 modelU = {"BASE_VALUE":1.0,"conts":{"x":[[0,103], [7,103]]},'featcomb':'mult'}
 modelP = {"BASE_VALUE":1.0,"conts":{"x":[[0,0.2], [7,0.2]]},'featcomb':'mult'}
 cdf = df[df['censored']].reset_index()
 udf = df[~df['censored']].reset_index()
 
 print(len(cdf))
 print(len(udf))
 #assert(False)
 #newModels = train_models([udf], "y", 1000, [10,0.001], [modelU, modelP], lossgrads = [[calculus.gnormal_u_diff, calculus.gnormal_p_diff]], links=[calculus.Take0,calculus.Take1], linkgrads=[[calculus.js1, calculus.js0],[calculus.js0, calculus.js1]])
 
 newModels = train_models([udf, cdf], "y", 1000, [10,0.005], [modelU, modelP], lossgrads = [[calculus.gnormal_u_diff, calculus.gnormal_p_diff],[calculus.u_diff_censored,calculus.p_diff_censored]], links=[calculus.Take0,calculus.Take1], linkgrads=[[calculus.js1, calculus.js0],[calculus.js0, calculus.js1]])
 for newModel in newModels:
  misc.explain(newModel)









if False:#__name__ == '__main__':
 df = pd.DataFrame({"cat1":[True, True, True, False, False, False],'cat2':[True, False, True, False, True, False],"y":[2,1,2,0,1,0]})
 model = {"BASE_VALUE":1.0,"conts":{}, "featcomb":'addl', "cats":{"cat1":{"uniques":{True:1,False:1,},"OTHER":1}, "cat2":{"uniques":{True:1,False:1},"OTHER":1}}}
 model = train_model(df, 'y', 500, 0.1, model, lossgrad=calculus.Gauss_grad)
 print(model)
 


if False:#__name__ == '__main__':
 df = pd.DataFrame({"cat1":['a','a','a','b','b','b','q','q','q','q'],'cat2':['c','d','q','c','d','q','c','d','q','q'],"y":[1,2,3,4,5,6,7,8,9,10]})
 model = {"BASE_VALUE":1.0,"conts":{}, "cats":{"cat1":{"uniques":{"a":1,"b":1,},"OTHER":1}, "cat2":{"uniques":{"c":1,"d":1},"OTHER":1}}, 'catcats':{'cat1 X cat2':{'uniques':{'a':{'uniques':{'c':1,'d':1},'OTHER':1},'b':{'uniques':{'c':1,'d':1},'OTHER':1}},'OTHER':{'uniques':{'c':1,'d':1},'OTHER':1}}}, 'catconts':{}, 'contconts':{}}
 print(rele.produce_catcat_relevances(df['cat1'], df['cat2'], model["catcats"]["cat1 X cat2"]))
 
 df = pd.DataFrame({"cat1":['a','a','a','b','b','b','q','q','q','q'],'cat2':['c','d','q','c','d','q','c','d','q','q'],"y":[1,2,3,4,5,6,7,8,9,10]})
 model = {"BASE_VALUE":1.0,"conts":{}, "cats":{"cat1":{"uniques":{"a":1,"b":1,},"OTHER":1}, "cat2":{"uniques":{"c":1,"d":1},"OTHER":1}}, 'catcats':{'cat1 X cat2':{'uniques':{'a':{'uniques':{'c':1,'d':1},'OTHER':1},'b':{'uniques':{'c':1,'d':2},'OTHER':1}},'OTHER':{'uniques':{'c':1,'d':1},'OTHER':1}}}, 'catconts':{}, 'contconts':{}}
 reles = rele.produce_catcat_relevances(df['cat1'], df['cat2'], model["catcats"]["cat1 X cat2"])
 print(misc.get_effect_of_this_catcat_from_relevances(reles, model, "cat1 X cat2"))
 
 df = pd.DataFrame({"cat1":['a','a','a','b','b','b','q','q','q','q'],'cat2':['c','d','q','c','d','q','c','d','q','q'],"y":[1,1,1,1,2,1,1,1,1,1]})
 model = {"BASE_VALUE":1.0,"conts":{}, "cats":{"cat1":{"uniques":{"a":1,"b":1,},"OTHER":1}, "cat2":{"uniques":{"c":1,"d":1},"OTHER":1}}, 'catcats':{'cat1 X cat2':{'uniques':{'a':{'uniques':{'c':1,'d':1},'OTHER':1},'b':{'uniques':{'c':1,'d':1},'OTHER':1}},'OTHER':{'uniques':{'c':1,'d':1},'OTHER':1}}}, 'catconts':{}, 'contconts':{}}
 
 newModel = train_model(df, "y",50, 0.4, model, staticFeats = ['cat1','cat2'])
 print(newModel)
 
 
 df = pd.DataFrame({"cat1":['a','a','a','b','b','b','q','q','q','q','q'],'cont1':[1,2,3,1,2,3,1,2,3,1.5,np.nan],"y":[1,2,3,4,5,6,7,8,9,10,11]})
 model = {"BASE_VALUE":1.0,"conts":{'cont1':[[1,1],[2,1],[3,1]]}, "cats":{"cat1":{"uniques":{"a":1,"b":1},"OTHER":1}}, 'catcats':{}, 'catconts':{"cat1 X cont1":{"uniques":{"a":[[1,1],[2,1],[3,1]],"b":[[1,1],[2,1],[3,1]]},"OTHER":[[1,1],[2,1],[3,1]]}}, 'contconts':{}}
 print(rele.produce_catcont_relevances(df['cat1'], df['cont1'], model["catconts"]["cat1 X cont1"]))
 
 df = pd.DataFrame({"cat1":['a','a','a','b','b','b','q','q','q','q','q'],'cont1':[1,2,3,1,2,3,1,2,3,1.5,np.nan],"y":[1,2,3,4,5,6,7,8,9,10,11]})
 model = {"BASE_VALUE":1.0,"conts":{'cont1':[[1,1],[2,1],[3,1]]}, "cats":{"cat1":{"uniques":{"a":1,"b":1},"OTHER":1}}, 'catcats':{}, 'catconts':{"cat1 X cont1":{"uniques":{"a":[[1,1],[2,1],[3,1]],"b":[[1,1],[2,4],[3,1]]},"OTHER":[[1,1],[2,1],[3,1]]}}, 'contconts':{}}
 reles = rele.produce_catcont_relevances(df['cat1'], df['cont1'], model["catconts"]["cat1 X cont1"])
 print(misc.get_effect_of_this_catcont_from_relevances(reles, model, "cat1 X cont1"))
 
 df = pd.DataFrame({"cat1":['a','a','a','b','b','b','q','q','q','q','q'],'cont1':[1,2,3,1,2,3,1,2,3,1.5,np.nan],"y":[1,1,1,1,1,2,1,1,1,1,1]})
 model = {"BASE_VALUE":1.0,"conts":{'cont1':[[1,1],[2,1],[3,1]]}, "cats":{"cat1":{"uniques":{"a":1,"b":1},"OTHER":1}}, 'catcats':{}, 'catconts':{"cat1 X cont1":{"uniques":{"a":[[1,1],[2,1],[3,1]],"b":[[1,1],[2,1],[3,1]]},"OTHER":[[1,1],[2,1],[3,1]]}}, 'contconts':{}}
 newModel = train_model(df, "y",50, 0.4, model, staticFeats = ['cat1','cont1'])
 print(newModel)
 
 df = pd.DataFrame({"cont1":[1,1,1,2,2,2,3,3,3,1.5,np.nan],'cont2':[1,2,3,1,2,3,1,2,3,1.5,np.nan],"y":[1,2,3,4,5,6,7,8,9,10,11]})
 model = {"BASE_VALUE":1.0,"conts":{'cont1':[[1,1],[2,1],[3,1]],'cont2':[[1,1],[2,1],[3,1]]}, "cats":{}, 'catcats':{}, 'catconts':{}, 'contconts':{'cont1 X cont2': [[1,[[1,1],[2,1],[3,1]]],[2,[[1,1],[2,1],[3,1]]],[3,[[1,1],[2,1],[3,1]]]]} }
 print(rele.produce_contcont_relevances(df['cont1'], df['cont2'], model["contconts"]["cont1 X cont2"]))
 
 df = pd.DataFrame({"cont1":[1,1,1,2,2,2,3,3,3,1.5,np.nan],'cont2':[1,2,3,1,2,3,1,2,3,1.5,np.nan],"y":[1,2,3,4,5,6,7,8,9,10,11]})
 models = [{"BASE_VALUE":1.0,"conts":{'cont1':[[1,1],[2,1],[3,1]],'cont2':[[1,1],[2,1],[3,1]]}, "cats":{}, 'catcats':{}, 'catconts':{}, 'contconts':{'cont1 X cont2': [[1,[[1,1],[2,1],[3,1]]],[2,[[1,1],[2,1],[3,1]]],[3,[[1,1],[2,1],[3,5]]]]} }]
 reles = rele.produce_contcont_relevances(df['cont1'], df['cont2'], model["contconts"]["cont1 X cont2"])
 print(misc.get_effect_of_this_contcont_from_relevances(reles, models[0], "cont1 X cont2"))
 
 df = pd.DataFrame({"cont1":[1,1,1,2,2,2,3,3,3,1.5,np.nan],'cont2':[1,2,3,1,2,3,1,2,3,1.5,np.nan],"y":[1,1,1, 1,1,1, 1,1,2, 1,1]})
 model = {"BASE_VALUE":1.0,"conts":{'cont1':[[1,1],[2,1],[3,1]],'cont2':[[1,1],[2,1],[3,1]]}, "cats":{}, 'catcats':{}, 'catconts':{}, 'contconts':{'cont1 X cont2': [[1,[[1,1],[2,1],[3,1]]],[2,[[1,1],[2,1],[3,1]]],[3,[[1,1],[2,1],[3,1]]]]} }
 newModel = train_model(df, "y",50, 0.4, model, staticFeats = ['cont1','cont2'])
 print(newModel)
 
if False:
 df = pd.DataFrame({"x":[1,2,3,4,5,6,7,8,9],"y":[1,2,3,4,5,6,7,8,9]})
 models = [{"BASE_VALUE":1.0,"conts":{"x":[[1,5],[5,5],[9,5]]}, "cats":[]}]
 newModels = train_model(df, "y",100, [2], models,weights=[1,2,3,4,5,6,7,8,9])
 for newModel in newModels:
  misc.explain(newModel)


if False:
 df = pd.DataFrame({"x":[1,2,3,4,5,6,7,8,9],"y":[2,2,2,2,2,2,2,2,2]})
 models = [{"BASE_VALUE":1.0,"conts":{"x":[[2,1],[5,1],[8,1]]}, "cats":[]}]
 print(produce_cont_relevances(df, models[0], "x"))
 
 df = pd.DataFrame({"x":["cat","dog","cat","mouse","rat"],"y":[2,2,2,2,2]})
 models = [{"BASE_VALUE":1.0,"conts":{}, "cats":{"x":{"uniques":{"cat":1,"mouse":1,"dog":1},"OTHER":1}}}]
 print(produce_cat_relevances(df, models[0], "x"))
 
 df = pd.DataFrame({"x":[1,2,3,4,5,6,7,8,9],"y":[2,2,2,2,2,2,2,2,2]})
 models = [{"BASE_VALUE":1.0,"conts":{"x":[[2,1],[5,2],[8,1]]}, "cats":[]}]
 reles = produce_cont_relevances(df, models[0], "x")
 print(misc.get_effect_of_this_cont_col_from_relevances(reles, models[0], "x"))
 
 df = pd.DataFrame({"x":["cat","dog","cat","mouse","rat"],"y":[2,2,2,2,2]})
 models = [{"BASE_VALUE":1.0,"conts":{}, "cats":{"x":{"uniques":{"cat":2,"mouse":1,"dog":3},"OTHER":1.5}}}]
 reles = produce_cat_relevances(df, models[0], "x")
 print(misc.get_effect_of_this_cat_col_from_relevances(reles, models[0], "x"))
 
 df = pd.DataFrame({"x":[1,2,3],"y":[2,2,2]})
 models = [{"BASE_VALUE":1.0,"conts":{"x":[[0,1],[4,1]]}, "cats":[]}]
 newModels = train_model(df, "y",50, [2.0], models)
 for newModel in newModels:
  misc.explain(newModel)
 
 df = pd.DataFrame({"x":[1,2,3],"y":[2,3,4]})
 models = [{"BASE_VALUE":1.0,"conts":{"x":[[0,1],[4,1]]}, "cats":[]}]
 newModels = train_model(df, "y",100, [2.0], models)
 for newModel in newModels:
  misc.explain(newModel)
 
 df = pd.DataFrame({"x":[1,2],"y":[2,2]})
 models = [{"BASE_VALUE":1.0,"conts":{"x":[[1,1],[2,1]]}, "cats":[]}]
 newModels = train_model(df, "y",100, [2.0], models)
 for newModel in newModels:
  misc.explain(newModel)
 
 df = pd.DataFrame({"x":[1, 995,996,997,998,999,1000],"y":[2,2,2,2,2,2,2]})
 models = [{"BASE_VALUE":1.0,"conts":{"x":[[1,1],[1000,1]]}, "cats":[]}]
 newModels = train_model(df, "y",50, [1.5], models)
 for newModel in newModels:
  misc.explain(newModel)
 
 df = pd.DataFrame({"x":[1,2,3,4],"y":[1+1,2+1.2,3+1.4,4+1.6]})
 models = [{"BASE_VALUE":2,"conts":{"x":[[1,1],[4,1]]}, "cats":[]},{"BASE_VALUE":1,"conts":{"x":[[1,1],[4,1]]}, "cats":[]}]
 newModels = train_model(df, "y",100, [1.5, 1.5], models)
 for newModel in newModels:
  misc.explain(newModel)

 df = pd.DataFrame({"x":["cat","dog","cat","mouse","rat"],"y":[2.0,3.0,2.0,1.0,1.5]})
 models = [{"BASE_VALUE":1.0,"conts":{}, "cats":{"x":{"uniques":{"cat":1,"mouse":1,"dog":1},"OTHER":1}}}]
 newModels = train_model(df, "y",100, [0.5], models)
 for newModel in newModels:
  misc.explain(newModel)
 



import pandas as pd
import numpy as np

import copy
import os

from . import actual_modelling
from . import prep
from . import misc
from . import calculus
from . import viz

ALPHABET="ABCDEFGHIJKLMNOPQRSTUVWXYZ"

def prep_model(inputDf, resp, cats, conts, catMinPrev=0.01, contTargetPts=5, contEdge=0.01, weightCol=None):
 df = inputDf.reset_index(drop=True)
 model = prep.prep_model(df, resp, cats, conts, catMinPrev, contTargetPts, contEdge, 1, weightCol)
 return model

def prep_multiplicative_model(inputDf, resp, cats, conts, catMinPrev=0.01, contTargetPts=5, contEdge=0.01, weightCol=None):
 return prep_model(inputDf, resp, cats, conts, catMinPrev=0.01, contTargetPts=5, contEdge=0.01, weightCol=None)


def train_gamma_model(inputDf, resp, nrounds, lr, model, pen=0, weightCol=None, staticFeats=[], prints="normal"):
 df = inputDf.reset_index(drop=True)
 model = actual_modelling.train_model(df, resp, nrounds, lr, model, weightCol, staticFeats, pen=pen, specificPens={}, lossgrad=calculus.Gamma_grad, prints=prints)
 return model

def train_poisson_model(inputDf, resp, nrounds, lr, model, pen=0, weightCol=None, staticFeats=[], prints="normal"):
 df = inputDf.reset_index(drop=True)
 model = actual_modelling.train_model(df, resp, nrounds, lr, model, weightCol, staticFeats, pen=pen, specificPens={}, lossgrad=calculus.Poisson_grad, prints=prints)
 return model

def train_tweedie_model(inputDf, resp, nrounds, lr, model, pTweedie=1.5, pen=0, weightCol=None, staticFeats=[], prints="normal"):
 df = inputDf.reset_index(drop=True)
 Tweedie_grad = calculus.produce_Tweedie_grad(pTweedie) 
 model = actual_modelling.train_model(df, resp, nrounds, lr, model, weightCol, staticFeats, pen=pen, specificPens={}, lossgrad=Tweedie_grad, prints=prints)
 return model





def interxhunt_gamma_model(inputDf, resp, cats, conts, model, silent=False, weightCol=None, filename="suggestions"):
 
 df = inputDf.reset_index(drop=True)
 
 df["PredComb"]=misc.predict(df,model)
 
 sugImps=[]
 sugFeats=[]
 sugTypes=[]
 
 for i in range(len(cats)):
  for j in range(i+1, len(cats)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}}
   prep.add_catcat_to_model(trialmodel, df, cats[i], cats[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=calculus.Gamma_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+cats[j], misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=1))
   
   sugFeats.append(cats[i]+" X "+cats[j])
   sugImps.append(misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=1))
   sugTypes.append("catcat")
 
 for i in range(len(cats)):
  for j in range(len(conts)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}}
   prep.add_catcont_to_model(trialmodel, df, cats[i], conts[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=calculus.Gamma_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+conts[j],misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=1))
   
   sugFeats.append(cats[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=1))
   sugTypes.append("catcont")
 
 for i in range(len(conts)):
  for j in range(i+1, len(conts)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}}
   prep.add_contcont_to_model(trialmodel, df, conts[i], conts[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=calculus.Gamma_grad, pen=0, prints="silent")
   
   if not silent:
    print(conts[i]+" X "+conts[j], misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=1))
   
   sugFeats.append(conts[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=1))
   sugTypes.append("contcont")
 
 sugDf = pd.DataFrame({"Interaction":sugFeats, "Type":sugTypes, "Importance":sugImps})
 sugDf = sugDf.sort_values(['Importance'], ascending=False).reset_index()
 sugDf.to_csv(filename+".csv")


def interxhunt_poisson_model(inputDf, resp, cats, conts, model, silent=False, weightCol=None, filename="suggestions"):
 
 df = inputDf.reset_index(drop=True)
 
 df["PredComb"]=misc.predict(df,model)
 
 sugImps=[]
 sugFeats=[]
 sugTypes=[]
 
 for i in range(len(cats)):
  for j in range(i+1, len(cats)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}}
   prep.add_catcat_to_model(trialmodel, df, cats[i], cats[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=calculus.Poisson_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+cats[j], misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=1))
   
   sugFeats.append(cats[i]+" X "+cats[j])
   sugImps.append(misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=1))
   sugTypes.append("catcat")
 
 for i in range(len(cats)):
  for j in range(len(conts)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}}
   prep.add_catcont_to_model(trialmodel, df, cats[i], conts[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=calculus.Poisson_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+conts[j],misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=1))
   
   sugFeats.append(cats[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=1))
   sugTypes.append("catcont")
 
 for i in range(len(conts)):
  for j in range(i+1, len(conts)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}}
   prep.add_contcont_to_model(trialmodel, df, conts[i], conts[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=calculus.Poisson_grad, pen=0, prints="silent")
   
   if not silent:
    print(conts[i]+" X "+conts[j], misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=1))
   
   sugFeats.append(conts[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=1))
   sugTypes.append("contcont")
 
 sugDf = pd.DataFrame({"Interaction":sugFeats, "Type":sugTypes, "Importance":sugImps})
 sugDf = sugDf.sort_values(['Importance'], ascending=False).reset_index()
 sugDf.to_csv(filename+".csv")




def interxhunt_tweedie_model(inputDf, resp, cats, conts, model, pTweedie=1.5, silent=False, weightCol=None, filename="suggestions"):
 
 df = inputDf.reset_index(drop=True)
 
 df["PredComb"]=misc.predict(df,model)
 
 sugImps=[]
 sugFeats=[]
 sugTypes=[]
 
 Tweedie_grad = calculus.produce_Tweedie_grad(pTweedie) 
 
 for i in range(len(cats)):
  for j in range(i+1, len(cats)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}}
   prep.add_catcat_to_model(trialmodel, df, cats[i], cats[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=Tweedie_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+cats[j], misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=1))
   
   sugFeats.append(cats[i]+" X "+cats[j])
   sugImps.append(misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=1))
   sugTypes.append("catcat")
 
 for i in range(len(cats)):
  for j in range(len(conts)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}}
   prep.add_catcont_to_model(trialmodel, df, cats[i], conts[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=Tweedie_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+conts[j],misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=1))
   
   sugFeats.append(cats[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=1))
   sugTypes.append("catcont")
 
 for i in range(len(conts)):
  for j in range(i+1, len(conts)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}}
   prep.add_contcont_to_model(trialmodel, df, conts[i], conts[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=Tweedie_grad, pen=0, prints="silent")
   
   if not silent:
    print(conts[i]+" X "+conts[j], misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=1))
   
   sugFeats.append(conts[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=1))
   sugTypes.append("contcont")
 
 sugDf = pd.DataFrame({"Interaction":sugFeats, "Type":sugTypes, "Importance":sugImps})
 sugDf = sugDf.sort_values(['Importance'], ascending=False).reset_index()
 sugDf.to_csv(filename+".csv")




def prep_gamma_models(inputDf, resp, cats, conts, N=1, fractions=None, catMinPrev=0.01, contTargetPts=5, contEdge=0.01, weightCol=None):
 df = inputDf.reset_index(drop=True)
 if fractions==None:
  denom = N*(N+1)/2
  fractions = [(N-x)/denom for x in range(N)]
 models = []
 for fraction in fractions:
  model = prep.prep_model(df, resp, cats, conts, catMinPrev, contTargetPts, contEdge, 1, weightCol)
  model["BASE_VALUE"]*=fraction
  models.append(model)
 return models

def train_gamma_models(inputDf, resp, nrounds, lrs, models, pens=None, weightCol=None, staticFeats=[], prints="normal"):
 df = inputDf.reset_index(drop=True)
 models = actual_modelling.train_models([df], resp, nrounds, lrs, models, weightCol, staticFeats, lras=calculus.addsmoothing_LRAs, lossgrads=[[calculus.Gamma_grad]], links=[calculus.Add_mlink], linkgrads=[[calculus.Add_mlink_grad]*len(models)], pens=pens, prints=prints)
 return models

def interxhunt_gamma_models(inputDf, resp, cats, conts, models, silent=False, weightCol=None, filename="suggestions"):
 
 df = inputDf.reset_index(drop=True)
 
 trialModelTemplate = []
 
 for m in range(len(models)):
  df["PredComb_"+str(m)]=misc.predict(df, models[m])
  trialModelTemplate.append({"BASE_VALUE":1, "conts":{"PredComb_"+str(m):[[min(df["PredComb_"+str(m)]),min(df["PredComb_"+str(m)])],[max(df["PredComb_"+str(m)]),max(df["PredComb_"+str(m)])]]}, "featcomb":"mult"})
 
 sugImps=[[] for m in range(len(models))]
 sugFeats=[[] for m in range(len(models))]
 sugTypes=[[] for m in range(len(models))]
 
 for i in range(len(cats)):
  for j in range(i+1, len(cats)):
   if not silent:
    print(cats[i] + " X " + cats[j])
   
   trialModels = copy.deepcopy(trialModelTemplate)
   trialModels = [prep.add_catcat_to_model(trialModel, df, cats[i], cats[j], defaultValue=1) for trialModel in trialModels]
   trialModels = train_gamma_models(df, resp, 1, [1]*len(models), trialModels, weightCol, staticFeats=["PredComb_"+str(m) for m in range(len(models))], prints="silent")
   print(trialModels)
   
   for m in range(len(models)):
    sugFeats[m].append(cats[i]+" X "+cats[j])
    sugImps[m].append(misc.get_importance_of_this_catcat(df, trialModels[m], cats[i]+" X "+cats[j], defaultValue=0))
    sugTypes[m].append("catcat")
 
 for i in range(len(cats)):
  for j in range(len(conts)):
   if not silent:
    print(cats[i] + " X " + conts[j])
   
   trialModels = copy.deepcopy(trialModelTemplate)
   trialModels = [prep.add_catcont_to_model(trialModel, df, cats[i], conts[j], defaultValue=1) for trialModel in trialModels]
   trialModels = train_gamma_models(df, resp, 1, [1]*len(models), trialModels, weightCol, staticFeats=["PredComb_"+str(m) for m in range(len(models))], prints="silent")
   
   for m in range(len(models)):
    sugFeats[m].append(cats[i]+" X "+conts[j])
    sugImps[m].append(misc.get_importance_of_this_catcont(df, trialModels[m], cats[i]+" X "+conts[j], defaultValue=0))
    sugTypes[m].append("catcont")
   
 for i in range(len(conts)):
  for j in range(i+1, len(conts)):
   if not silent:
    print(conts[i] + " X " + conts[j])
   
   trialModels = copy.deepcopy(trialModelTemplate)
   trialModels = [prep.add_contcont_to_model(trialModel, df, conts[i], conts[j], defaultValue=1) for trialModel in trialModels]
   trialModels = train_gamma_models(df, resp, 1, [1]*len(models), trialModels, weightCol, staticFeats=["PredComb_"+str(m) for m in range(len(models))], prints="silent")
   
   for m in range(len(models)):
    sugFeats[m].append(conts[i]+" X "+conts[j])
    sugImps[m].append(misc.get_importance_of_this_contcont(df, trialModels[m], conts[i]+" X "+conts[j], defaultValue=0))
    sugTypes[m].append("contcont")
 
 
 for m in range(len(models)):
  print(sugFeats[m],sugTypes[m],sugImps[m])
  sugDf = pd.DataFrame({"Interaction":sugFeats[m], "Type":sugTypes[m], "Importance":sugImps[m]})
  sugDf = sugDf.sort_values(['Importance'], ascending=False).reset_index()
  sugDf.to_csv(filename+"_"+ALPHABET[m]+".csv")
  

def gnormalize_gamma_models(models, inputDf, resp, cats, conts, startingErrorPercent=20, catMinPrev=0.01, contTargetPts=5, contEdge=0.01, weightCol=None):
 
 df = inputDf.reset_index(drop=True)
 
 errModel = prep.prep_model(df, resp, cats, conts, catMinPrev, contTargetPts, contEdge, 1, weightCol)
 errModel["BASE_VALUE"]=startingErrorPercent*1.25/100
 models.append(errModel)
 return models

def train_gnormal_models(inputDfs, resp, nrounds, lrs, models, pens=None, weightCol=None, staticFeats=[], prints="normal"):
 if type(inputDfs)!=list:
  dfs=[inputDfs.reset_index(drop=True)]
 else:
  dfs = [inputDf.reset_index(drop=True) for inputDf in inputDfs]
 
 models = actual_modelling.train_models(dfs, resp, nrounds, lrs, models, weightCol, staticFeats, lras = calculus.addsmoothing_LRAs_erry[:len(models)-1] + [calculus.default_LRA], lossgrads = [[calculus.gnormal_u_diff, calculus.gnormal_p_diff],[calculus.gnormal_u_diff_censored, calculus.gnormal_p_diff_censored]], links=[calculus.Add_mlink_allbutlast, calculus.Add_mlink_onlylast], linkgrads=[[calculus.Add_mlink_grad]*(len(models)-1)+[calculus.Add_mlink_grad_void], [calculus.Add_mlink_grad_void]*(len(models)-1)+[calculus.Add_mlink_grad]], pens=pens, prints=prints)
 
 return models

def interxhunt_gnormal_models(inputDfs, resp, cats, conts, models, silent=False, weightCol=None, filename="suggestions"):
 
 if type(inputDfs)!=list:
  dfs = [inputDfs.reset_index(drop=True)]
 else:
  dfs = [inputDf.reset_index(drop=True) for inputDf in inputDfs]
 
 if len(dfs)==2:
  cdf = dfs[0].append(dfs[1]).reset_index()
 else:
  cdf = dfs[0]
 
 trialModelTemplate = []
 
 
 for m in range(len(models)):
  for df in dfs:
   df["PredComb_"+str(m)]=misc.predict(df, models[m])
  
  minever = min([min(df["PredComb_"+str(m)]) for df in dfs])
  maxever = max([max(df["PredComb_"+str(m)]) for df in dfs])
  trialModelTemplate.append({"BASE_VALUE":1, "conts":{"PredComb_"+str(m):[[minever, minever], [maxever, maxever]]}, "featcomb":"mult"})
 
 
 sugImps=[[] for m in range(len(models))]
 sugFeats=[[] for m in range(len(models))]
 sugTypes=[[] for m in range(len(models))]
 
 for i in range(len(cats)):
  for j in range(i+1, len(cats)):
   if not silent:
    print(cats[i] + " X " + cats[j])
   
   trialModels = copy.deepcopy(trialModelTemplate)
   trialModels = [prep.add_catcat_to_model(trialModel, cdf, cats[i], cats[j], defaultValue=1) for trialModel in trialModels]
   trialModels = train_gnormal_models(dfs, resp, 1, [1]*len(models), trialModels, weightCol, staticFeats=["PredComb_"+str(m) for m in range(len(models))], prints="silent")
   
   for m in range(len(models)):
    sugFeats[m].append(cats[i]+" X "+cats[j])
    sugImps[m].append(misc.get_importance_of_this_catcat(cdf, trialModels[m], cats[i]+" X "+cats[j], defaultValue=0))
    sugTypes[m].append("catcat")
 
 for i in range(len(cats)):
  for j in range(len(conts)):
   if not silent:
    print(cats[i] + " X " + conts[j])
   
   trialModels = copy.deepcopy(trialModelTemplate)
   trialModels = [prep.add_catcont_to_model(trialModel, cdf, cats[i], conts[j], defaultValue=1) for trialModel in trialModels]
   trialModels = train_gnormal_models(dfs, resp, 1, [1]*len(models), trialModels, weightCol, staticFeats=["PredComb_"+str(m) for m in range(len(models))], prints="silent")
   
   for m in range(len(models)):
    sugFeats[m].append(cats[i]+" X "+conts[j])
    sugImps[m].append(misc.get_importance_of_this_catcont(cdf, trialModels[m], cats[i]+" X "+conts[j], defaultValue=0))
    sugTypes[m].append("catcont")
   
 for i in range(len(conts)):
  for j in range(i+1, len(conts)):
   if not silent:
    print(conts[i] + " X " + conts[j])
   
   trialModels = copy.deepcopy(trialModelTemplate)
   trialModels = [prep.add_contcont_to_model(trialModel, cdf, conts[i], conts[j], defaultValue=1) for trialModel in trialModels]
   trialModels = train_gnormal_models(dfs, resp, 1, [1]*len(models), trialModels, weightCol, staticFeats=["PredComb_"+str(m) for m in range(len(models))], prints="silent")
   
   for m in range(len(models)):
    sugFeats[m].append(conts[i]+" X "+conts[j])
    sugImps[m].append(misc.get_importance_of_this_contcont(cdf, trialModels[m], conts[i]+" X "+conts[j], defaultValue=0))
    sugTypes[m].append("contcont")
 
 
 for m in range(len(models)-1):
  sugDf = pd.DataFrame({"Interaction":sugFeats[m], "Type":sugTypes[m], "Importance":sugImps[m]})
  sugDf = sugDf.sort_values(['Importance'], ascending=False).reset_index()
  sugDf.to_csv(filename+"_"+ALPHABET[m]+".csv")


def predict_from_gnormal(df, model):
 return misc.predict_models(df, model, calculus.Add_mlink_allbutlast)
 
def predict_error_from_gnormal(df, model):
 return misc.predict_models(df, model, calculus.Add_mlink_onlylast)*100/1.25


def prep_additive_model(inputDf, resp, cats, conts, catMinPrev=0.01, contTargetPts=5, contEdge=0.01, weightCol=None):
 df = inputDf.reset_index(drop=True)
 model = prep.prep_model(df, resp, cats, conts, catMinPrev, contTargetPts, contEdge, 0, weightCol)
 model["featcomb"] = "addl"
 return model

def train_normal_model(inputDf, resp, nrounds, lr, model, pen=0, weightCol=None, staticFeats=[], prints="normal"):
 df = inputDf.reset_index(drop=True)
 model = actual_modelling.train_model(df, resp, nrounds, lr, model, weightCol, staticFeats, pen=pen, specificPens={}, lossgrad=calculus.Gauss_grad, prints=prints)
 return model

def interxhunt_normal_model(inputDf, resp, cats, conts, model, silent=False, weightCol=None, filename="suggestions"):
 
 df = inputDf.reset_index(drop=True)
 
 df["PredComb"]=misc.predict(df,model)
 
 sugImps=[]
 sugFeats=[]
 sugTypes=[]
 
 for i in range(len(cats)):
  for j in range(i+1, len(cats)):
   trialmodel = {"BASE_VALUE":0, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}, "featcomb":"addl"}
   prep.add_catcat_to_model(trialmodel, df, cats[i], cats[j], defaultValue=0)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=calculus.Gauss_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+cats[j], misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=0))
   
   sugFeats.append(cats[i]+" X "+cats[j])
   sugImps.append(misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=0))
   sugTypes.append("catcat")
 
 for i in range(len(cats)):
  for j in range(len(conts)):
   trialmodel = {"BASE_VALUE":0, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}, "featcomb":"addl"}
   prep.add_catcont_to_model(trialmodel, df, cats[i], conts[j], defaultValue=0)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=calculus.Gauss_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+conts[j],misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=0))
   
   sugFeats.append(cats[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=0))
   sugTypes.append("catcont")
 
 for i in range(len(conts)):
  for j in range(i+1, len(conts)):
   trialmodel = {"BASE_VALUE":0, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}, "featcomb":"addl"}
   prep.add_contcont_to_model(trialmodel, df, conts[i], conts[j], defaultValue=0)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad=calculus.Gauss_grad, pen=0, prints="silent")
   
   if not silent:
    print(conts[i]+" X "+conts[j], misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=0))
   
   sugFeats.append(conts[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=0))
   sugTypes.append("contcont")
 
 sugDf = pd.DataFrame({"Interaction":sugFeats, "Type":sugTypes, "Importance":sugImps})
 sugDf = sugDf.sort_values(['Importance'], ascending=False).reset_index()
 sugDf.to_csv(filename+".csv")


def prep_classifier_model(inputDf, resp, cats, conts, catMinPrev=0.01, contTargetPts=5, contEdge=0.01, weightCol=None):
 
 df = inputDf.reset_index(drop=True)
 
 model = prep.prep_model(df, resp, cats, conts, catMinPrev, contTargetPts, contEdge, 0, weightCol)
 model["BASE_VALUE"] = calculus.Logit_delink(model["BASE_VALUE"])
 model["featcomb"] = "addl"
 return model

def train_classifier_model(inputDf, resp, nrounds, lr, model, pen=0, weightCol=None, staticFeats=[], prints="normal"):
 
 df = inputDf.reset_index(drop=True)
 
 model = actual_modelling.train_model(df, resp, nrounds, lr, model, weightCol, staticFeats, pen=pen, specificPens={}, lossgrad=calculus.Logistic_grad, link = calculus.Logit_link, linkgrad = calculus.Logit_link_grad, prints=prints)
 return model

def interxhunt_classifier_model(inputDf, resp, cats, conts, model, silent=False, weightCol=None, filename="suggestions"):
 
 df = inputDf.reset_index(drop=True)
 
 df["PredComb"]=misc.predict(df,model)
 
 sugImps=[]
 sugFeats=[]
 sugTypes=[]
 
 for i in range(len(cats)):
  for j in range(i+1, len(cats)):
   trialmodel = {"BASE_VALUE":0, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}, "featcomb":"addl"}
   prep.add_catcat_to_model(trialmodel, df, cats[i], cats[j], defaultValue=0)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], link = calculus.Logit_link,  linkgrad = calculus.Logit_link_grad, lossgrad = calculus.Logistic_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+cats[j], misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=0))
   
   sugFeats.append(cats[i]+" X "+cats[j])
   sugImps.append(misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=0))
   sugTypes.append("catcat")
 
 for i in range(len(cats)):
  for j in range(len(conts)):
   trialmodel = {"BASE_VALUE":0, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}, "featcomb":"addl"}
   prep.add_catcont_to_model(trialmodel, df, cats[i], conts[j], defaultValue=0)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], link = calculus.Logit_link, linkgrad = calculus.Logit_link_grad, lossgrad = calculus.Logistic_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+conts[j],misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=0))
   
   sugFeats.append(cats[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=0))
   sugTypes.append("catcont")
 
 for i in range(len(conts)):
  for j in range(i+1, len(conts)):
   trialmodel = {"BASE_VALUE":0, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}, "featcomb":"addl"}
   prep.add_contcont_to_model(trialmodel, df, conts[i], conts[j], defaultValue=0)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], link = calculus.Logit_link, linkgrad = calculus.Logit_link_grad, lossgrad = calculus.Logistic_grad, pen=0, prints="silent")
   
   if not silent:
    print(conts[i]+" X "+conts[j], misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=0))
   
   sugFeats.append(conts[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=0))
   sugTypes.append("contcont")
 
 sugDf = pd.DataFrame({"Interaction":sugFeats, "Type":sugTypes, "Importance":sugImps})
 sugDf = sugDf.sort_values(['Importance'], ascending=False).reset_index()
 sugDf.to_csv(filename+".csv")




def prep_adjustment_model(inputDf, resp, startingPoint, cats, conts, catMinPrev=0.01, contTargetPts=5, contEdge=0.01, weightCol=None):
 
 df = inputDf.reset_index(drop=True)
 
 model = prep.prep_model(df, resp, cats, conts, catMinPrev, contTargetPts, contEdge, 1, weightCol)
 model["BASE_VALUE"]=1
 model["featcomb"]="mult"
 if "conts" not in model:
  model["conts"]={}
 model["conts"][startingPoint] = [[min(df[startingPoint]), min(df[startingPoint])], [max(df[startingPoint]), max(df[startingPoint])]]
 return model

def train_adjustment_model(inputDf, resp, startingPoint, nrounds, lr, model, pen=0, weightCol=None, staticFeats=[], prints="normal"):
 
 df = inputDf.reset_index(drop=True)
 
 model = actual_modelling.train_model(df, resp, nrounds, lr, model, weightCol, staticFeats=[startingPoint]+staticFeats, pen=pen, specificPens={}, lossgrad=calculus.Gamma_grad, prints=prints)
 return model

def interxhunt_adjustment_model(inputDf, resp, cats, conts, model, silent=False, weightCol=None, filename="suggestions"):
 
 df = inputDf.reset_index(drop=True)
 
 df["PredComb"]=misc.predict(df,model)
 
 sugImps=[]
 sugFeats=[]
 sugTypes=[]
 
 for i in range(len(cats)):
  for j in range(i+1, len(cats)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}, "featcomb":"mult"}
   prep.add_catcat_to_model(trialmodel, df, cats[i], cats[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad = calculus.Gamma_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+cats[j], misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=0))
   
   sugFeats.append(cats[i]+" X "+cats[j])
   sugImps.append(misc.get_importance_of_this_catcat(df, trialmodel, cats[i]+" X "+cats[j], defaultValue=0))
   sugTypes.append("catcat")
 
 for i in range(len(cats)):
  for j in range(len(conts)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}, "featcomb":"mult"}
   prep.add_catcont_to_model(trialmodel, df, cats[i], conts[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad = calculus.Gamma_grad, pen=0, prints="silent")
   
   if not silent:
    print(cats[i]+" X "+conts[j],misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=0))
   
   sugFeats.append(cats[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_catcont(df, trialmodel, cats[i]+" X "+conts[j], defaultValue=0))
   sugTypes.append("catcont")
 
 for i in range(len(conts)):
  for j in range(i+1, len(conts)):
   trialmodel = {"BASE_VALUE":1, "conts":{"PredComb":[[min(df["PredComb"]),min(df["PredComb"])],[max(df["PredComb"]),max(df["PredComb"])]]}, "featcomb":"mult"}
   prep.add_contcont_to_model(trialmodel, df, conts[i], conts[j], defaultValue=1)
   trialmodel = actual_modelling.train_model(df, resp, 1, 0.2, trialmodel, staticFeats=["PredComb"], lossgrad = calculus.Gamma_grad, pen=0, prints="silent")
   
   if not silent:
    print(conts[i]+" X "+conts[j], misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=0))
   
   sugFeats.append(conts[i]+" X "+conts[j])
   sugImps.append(misc.get_importance_of_this_contcont(df, trialmodel, conts[i]+" X "+conts[j], defaultValue=0))
   sugTypes.append("contcont")
 
 sugDf = pd.DataFrame({"Interaction":sugFeats, "Type":sugTypes, "Importance":sugImps})
 sugDf = sugDf.sort_values(['Importance'], ascending=False).reset_index()
 sugDf.to_csv(filename+".csv")





def prep_cratio_models(inputDf, resp, cats, conts, N=1, fractions=None, catMinPrev=0.01, contTargetPts=5, contEdge=0.01, weightCol=None):
 df = inputDf.reset_index(drop=True)
 if fractions==None:
  denom = N*(N+1)/2
  fractions = [(N-x)/denom for x in range(N)]
 models = []
 for fraction in fractions:
  model = prep.prep_model(df, resp, cats, conts, catMinPrev, contTargetPts, contEdge, 1, weightCol)
  model["BASE_VALUE"] = misc.frac_to_ratio(model["BASE_VALUE"])
  model["BASE_VALUE"]*=fraction
  models.append(model)
 return models

def train_cratio_models(inputDf, resp, nrounds, lrs, models, pens=None, weightCol=None, staticFeats=[], minRela=0.1, prints="normal"):
 df = inputDf.reset_index(drop=True)
 models = actual_modelling.train_models([df], resp, nrounds, lrs, models, weightCol, staticFeats, lras=calculus.addsmoothing_LRAs, lossgrads=[[calculus.Logistic_grad]], links=[calculus.Cratio_mlink], linkgrads=[[calculus.Cratio_mlink_grad]*len(models)], pens=pens, minRela=minRela, prints=prints)
 return models

def interxhunt_cratio_models(inputDf, resp, cats, conts, models, silent=False, weightCol=None, filename="suggestions"):
 
 df = inputDf.reset_index(drop=True)
 
 trialModelTemplate = []
 
 for m in range(len(models)):
  df["PredComb_"+str(m)]=misc.predict(df, models[m])
  trialModelTemplate.append({"BASE_VALUE":1, "conts":{"PredComb_"+str(m):[[min(df["PredComb_"+str(m)]),min(df["PredComb_"+str(m)])],[max(df["PredComb_"+str(m)]),max(df["PredComb_"+str(m)])]]}, "featcomb":"mult"})
 
 sugImps=[[] for m in range(len(models))]
 sugFeats=[[] for m in range(len(models))]
 sugTypes=[[] for m in range(len(models))]
 
 for i in range(len(cats)):
  for j in range(i+1, len(cats)):
   if not silent:
    print(cats[i] + " X " + cats[j])
   
   trialModels = copy.deepcopy(trialModelTemplate)
   trialModels = [prep.add_catcat_to_model(trialModel, df, cats[i], cats[j], defaultValue=1) for trialModel in trialModels]
   trialModels = train_cratio_models(df, resp, 1, [1]*len(models), trialModels, weightCol, staticFeats=["PredComb_"+str(m) for m in range(len(models))], prints="silent")
   print(trialModels)
   
   for m in range(len(models)):
    sugFeats[m].append(cats[i]+" X "+cats[j])
    sugImps[m].append(misc.get_importance_of_this_catcat(df, trialModels[m], cats[i]+" X "+cats[j], defaultValue=0))
    sugTypes[m].append("catcat")
 
 for i in range(len(cats)):
  for j in range(len(conts)):
   if not silent:
    print(cats[i] + " X " + conts[j])
   
   trialModels = copy.deepcopy(trialModelTemplate)
   trialModels = [prep.add_catcont_to_model(trialModel, df, cats[i], conts[j], defaultValue=1) for trialModel in trialModels]
   trialModels = train_cratio_models(df, resp, 1, [1]*len(models), trialModels, weightCol, staticFeats=["PredComb_"+str(m) for m in range(len(models))], prints="silent")
   
   for m in range(len(models)):
    sugFeats[m].append(cats[i]+" X "+conts[j])
    sugImps[m].append(misc.get_importance_of_this_catcont(df, trialModels[m], cats[i]+" X "+conts[j], defaultValue=0))
    sugTypes[m].append("catcont")
   
 for i in range(len(conts)):
  for j in range(i+1, len(conts)):
   if not silent:
    print(conts[i] + " X " + conts[j])
   
   trialModels = copy.deepcopy(trialModelTemplate)
   trialModels = [prep.add_contcont_to_model(trialModel, df, conts[i], conts[j], defaultValue=1) for trialModel in trialModels]
   trialModels = train_cratio_models(df, resp, 1, [1]*len(models), trialModels, weightCol, staticFeats=["PredComb_"+str(m) for m in range(len(models))], prints="silent")
   
   for m in range(len(models)):
    sugFeats[m].append(conts[i]+" X "+conts[j])
    sugImps[m].append(misc.get_importance_of_this_contcont(df, trialModels[m], conts[i]+" X "+conts[j], defaultValue=0))
    sugTypes[m].append("contcont")





def flinkify_additive_model(inputDf, model, contTargetPts=5, contEdge=0.0, weightCol=None):
 df = inputDf.reset_index(drop=True)
 df["preds"] = misc.predict(df, model)
 model["flink"] = prep.get_cont_feat(df, "preds",contTargetPts, contEdge, 0, weightCol=weightCol)
 return model
 
def flinkify_multiplicative_model(inputDf, model, contTargetPts=5, contEdge=0.0, weightCol=None):
 df = inputDf.reset_index(drop=True)
 df["preds"] = misc.predict(df, model)
 model["flink"] = prep.get_cont_feat(df, "preds",contTargetPts, contEdge, 1, weightCol=weightCol)
 return model


def viz_model(model, modelName=0, targetSpan=0.5, defaultValue=1, ytitle="Relativity", subfolder=None, otherName="OTHER", heatmapDetail=10):
 
 try:
  os.mkdir("graphs")
 except:
  pass
 
 if subfolder!=None:
  try:
   os.mkdir("graphs/"+subfolder)
  except:
   pass
  
  fsf = "graphs/"+subfolder
 else:
  fsf="graphs"
 
 
 if "flink" in model:
  viz.draw_cont_pdp(model["flink"], targetSpan, "Un-Flinked Prediction", model=modelName, defaultValue=defaultValue, ytitle=ytitle, folder=fsf)
 if "conts" in model:
  for col in model["conts"]:
   viz.draw_cont_pdp(model["conts"][col], targetSpan, col, model=modelName, defaultValue=defaultValue, ytitle=ytitle, folder=fsf)
 if "cats" in model:
  for col in model["cats"]:
   viz.draw_cat_pdp(model["cats"][col], targetSpan, col, model=modelName, defaultValue=defaultValue, ytitle=ytitle, folder=fsf, otherName=otherName)
 if "contconts" in model:
  for cols in model["contconts"]:
   c1, c2 = cols.split(" X ")
   viz.draw_contcont_pdp(model["contconts"][cols], targetSpan, cols, model=modelName, cont1=c1, cont2=c2, defaultValue=defaultValue, ytitle=ytitle, folder=fsf)
   viz.draw_contcont_pdp_3D(model["contconts"][cols], targetSpan, cols+", 3D", model=modelName, cont1=c1, defaultValue=defaultValue, cont2=c2, ytitle=ytitle, folder=fsf)
   viz.draw_contcont_pdp_heatmap(model["contconts"][cols], targetSpan, cols+", Heatmap", model=modelName, cont1=c1, defaultValue=defaultValue, cont2=c2, ytitle=ytitle, folder=fsf, detail=heatmapDetail)
 if "catconts" in model:
  for cols in model["catconts"]:
   c1, c2 = cols.split(" X ")
   viz.draw_catcont_pdp(model["catconts"][cols], targetSpan, cols, model=modelName, cat=c1, cont=c2, defaultValue=defaultValue, ytitle=ytitle, folder=fsf, otherName=otherName)
 if "catcats" in model:
  for cols in model["catcats"]:
   c1, c2 = cols.split(" X ")
   viz.draw_catcat_pdp(model["catcats"][cols], targetSpan, cols, model=modelName, cat1=c1, cat2=c2, defaultValue=defaultValue, ytitle=ytitle, folder=fsf, otherName=otherName)

def viz_logistic_model(model, subfolder=None, targetSpan=0.5, otherName="OTHER"):
 viz_model(model, defaultValue=0, ytitle="LPUs", subfolder=subfolder, targetSpan=targetSpan, otherName=otherName)

def viz_gamma_models(models, subfolder=None, targetSpan=0.5, otherName="OTHER"):
 if len(models)==1:
  viz_model(models[0], subfolder=subfolder, targetSpan=targetSpan, otherName=otherName)
 else:
  for m in range(len(models)):
   viz_model(models[m], modelName=ALPHABET[m], subfolder=subfolder, targetSpan=targetSpan, otherName=otherName)

def viz_gnormal_models(models, subfolder=None, targetSpan=0.5, otherName="OTHER"):
 if len(models)==2:
  viz_model(models[0], subfolder=subfolder, targetSpan=targetSpan, otherName=otherName)
 else:
  for m in range(len(models)-1):
   viz_model(models[m], modelName=ALPHABET[m], subfolder=subfolder, targetSpan=targetSpan, otherName=otherName)
 viz_model(models[-1], modelName="PercentageError", subfolder=subfolder, targetSpan=targetSpan, otherName=otherName)

def viz_additive_model(model, subfolder=None, targetSpan=0.5, otherName="OTHER"):
 viz_model(model, defaultValue=0, ytitle="Delta", subfolder=subfolder, targetSpan=targetSpan, otherName=otherName)

def viz_multiplicative_model(model, subfolder=None, targetSpan=0.5, otherName="OTHER"):
 viz_model(model, defaultValue=1, ytitle="Multiplier", subfolder=subfolder, targetSpan=targetSpan, otherName=otherName)

def viz_adjustment_model(model, subfolder=None, targetSpan=0.5, otherName="OTHER"):
 viz_model(model, ytitle="Adjustment multiplier", subfolder=subfolder, targetSpan=targetSpan, otherName=otherName)
 
def viz_cratio_models(models, subfolder=None, targetSpan=0.5, otherName="OTHER"):
 if len(models)==1:
  viz_model(models[0], subfolder=subfolder, targetSpan=targetSpan, otherName=otherName)
 else:
  for m in range(len(models)):
   viz_model(models[m], modelName=ALPHABET[m], subfolder=subfolder, targetSpan=targetSpan, otherName=otherName)



import numpy as np
import pandas as pd

def move_to_default(x, l, defaultValue=1):
 if x>defaultValue:
  x=max(x-l, defaultValue)
 elif x<defaultValue:
  x=min(x+l, defaultValue)
 return x

def penalize_model(model, pen, defaultValue=1, specificPenas={}):
 
 if "cats" in model:
  for cat in model["cats"]:
  
   if cat in specificPenas:
    l = specificPenas[cat]
   else:
    l = pen
   
   for u in model["cats"][cat]["uniques"]:
    model["cats"][cat]["uniques"][u] = move_to_default(model["cats"][cat]["uniques"][u], l, defaultValue)
   model["cats"][cat]["OTHER"] = move_to_default(model["cats"][cat]["OTHER"], l, defaultValue)
 
 if "conts" in model:
  for cont in model["conts"]:
   
   if cont in specificPenas:
    l = specificPenas[cont]
   else:
    l = pen
   
   for i in range(len(model["conts"][cont])):
    model["conts"][cont][i][1] = move_to_default(model["conts"][cont][i][1], l, defaultValue)
 
 if "catcats" in model:
  for catcat in model["catcats"]:
   
   if catcat in specificPenas:
    l = specificPenas[catcat]
   else:
    l = pen
   
   for u1 in model["catcats"][catcat]["uniques"]:
    model["catcats"][catcat]["uniques"][u1]["OTHER"] = move_to_default(model["catcats"][catcat]["uniques"][u1]["OTHER"], l, defaultValue)
    for u2 in model["catcats"][catcat]["uniques"][u1]["uniques"]:
     model["catcats"][catcat]["uniques"][u1]["uniques"][u2] = move_to_default(model["catcats"][catcat]["uniques"][u1]["uniques"][u2], l, defaultValue)
   for u2 in model["catcats"][catcat]["OTHER"]["uniques"]:
    model["catcats"][catcat]["OTHER"]["uniques"][u2] = move_to_default(model["catcats"][catcat]["OTHER"]["uniques"][u2], l, defaultValue)
   model["catcats"][catcat]["OTHER"]["OTHER"]=move_to_default(model["catcats"][catcat]["OTHER"]["OTHER"], l, defaultValue)
 
 if "catconts" in model:
  for catcont in model["catconts"]:
   
   if catcont in specificPenas:
    l = specificPenas[catcont]
   else:
    l = pen
   
   for u in model["catconts"][catcont]["uniques"]:
    for i in range(len(model["catconts"][catcont]["uniques"][u])):
     model["catconts"][catcont]["uniques"][u][i][1] = move_to_default(model["catconts"][catcont]["uniques"][u][i][1], l, defaultValue)
   for i in range(len(model["catconts"][catcont]["OTHER"])):
    model["catconts"][catcont]["OTHER"][i][1] = move_to_default(model["catconts"][catcont]["OTHER"][i][1], l, defaultValue)
 
 if "contconts" in model:
  for contcont in model["contconts"]:
   
   if contcont in specificPenas:
    l = specificPenas[contcont]
   else:
    l = pen
   
   for i in range(len(model["contconts"][contcont])):
    for j in range(len(model["contconts"][contcont][i][1])):
     model["contconts"][contcont][i][1][j][1] = move_to_default(model["contconts"][contcont][i][1][j][1], l, defaultValue)
 
 return model

if __name__ == '__main__':
 model={"BASE_VALUE":4, "conts":{"cont1":[[1,1.4], [2,0.7]]}, "cats":{"cat1":{"OTHER":1.9, "uniques":{"a":1.05, "b":0.2}}}, "catcats":{"cat1 X cat2": {"OTHER":{"OTHER":1.1, "uniques":{"c":0.99, "d":0.21}}, "uniques":{"a":{"OTHER":1.1, "uniques":{"c":0.99, "d":0.21}}, "b":{"OTHER":1.1, "uniques":{"c":0.99, "d":0.21}}}}}}
 model = penalize_model(model, 0.1)
 print(model)
 
 